#include "gen_sys.h"

//global includes
#include <iostream>
#include <string>
#include <math.h>
#include <stdio.h>
#include <fstream>
#include <list>
#include <map>

//local includes
#include "randomc.h"
#include "../linux_port/graph.h"
#include "../linux_port/system.h"

using namespace std;

//static helper functions

//Generate num_procs processors that have a bound from min_bound to max_bound
//and add them to the system, a bus is also connected to each of these processors and each
//processor is assigned the same OS
static void GenerateProcs(int num_procs, float min_bound, float max_bound);

//Generate num_comps components with a utilization between max_util and min_util with the
//given period, use_GME_hw is passed along.  Components will stop being generated if
//their total utilization becomes within 95% of the total bound for each processor (this
//is so their is a better possibility that the system is schedulable)
static void GenerateComponents(int num_comps, float max_util, float min_util, double period, bool use_GME_hw);

//Take the components that were generated in the GenerateComponents function and put
//num_comps of these components into the given set of tasks and remove them from the
//connecting_tasks set (see connecting_tasks definition below)
static void AssignIOComponents(int num_comps, map<int, TaskType*>& tasks);

//Remove the empty slots from the given task map i.e if there is an index i s.t i is
//not the last index there must be a valid index i+1
static void FillInHoles(map<int, TaskType*>& tasks);

//Generate connections between tasks in the begining tasks to any task in the connecting
//tasks or output tasks and between tasks in the connecting tasks and any task with an 
//index higher than it or a task in the output tasks.  Each task can is bounded by the
//number of out connections by the min and max output_degree values and the connection
//has a message size between the min and max msg_size
static void GenerateConnections(int max_output_degree, int min_output_degree, int max_msg_size, int min_msg_size);

//Store the descriptions of the given system and graph into a file
void StoreSystem(System& system, Graph& graph);

//Returns the sum of all of the bounds of the processors in the rand_sys if use_GME_hw is false
//otherwise it uses the processors in GME_sys
static double GetTotalBound(bool use_GME_hw);

//Delete and clear all values from the rand_sys and rand_grp 
static void CleanUp();

//static globals
static map<int, TaskType*> input_tasks;
static map<int, TaskType*> connecting_tasks;
static map<int, TaskType*> output_tasks;

//externals
extern Graph  rand_grp;
extern System rand_sys;
extern System GME_sys;
extern double g_bus_speed;	

void GenerateRandSys(RandSysParams param)
{
	SeedRandomGen(time(NULL));		//seed the random generator with the time

	input_tasks.clear();			//clear out lists
	connecting_tasks.clear();
	output_tasks.clear();

	rand_grp.SetDescription("");	//if there is a description clear it out
	rand_sys.SetDescription("");

	if(!param.append_sys)			//not adding to the system so clear it out
		CleanUp();

	int num_procs  = Random(param.min_procs,			param.max_procs);
	int num_comps  = Random(param.min_comps,			param.max_comps);
	double period  = Random(param.min_period,			param.max_period);
	int inp_comps  = Random(param.min_input_comps,		param.max_input_comps);
	int out_comps  = Random(param.min_output_comps,		param.max_output_comps);
	int out_degree = Random(param.min_output_degree,	param.max_output_degree);

	//only generate processors if we are not using GME hardware and we are not adding
	//to a current system
	if(!param.use_GME_hw)
		GenerateProcs(num_procs, param.max_bound, param.min_bound);

	GenerateComponents(num_comps, param.max_util, param.min_util, period, param.use_GME_hw);

	if(connecting_tasks.size() < 2)
		return;

	//input and output components can not be more than 1/4 the total number of tasks
	//but they must be at least 1
	if(inp_comps > connecting_tasks.size()/4)	inp_comps = connecting_tasks.size()/4;
	if(out_comps > connecting_tasks.size()/4)	out_comps = connecting_tasks.size()/4;
	if(inp_comps == 0)							inp_comps = 1;
	if(out_comps == 0)							inp_comps = 1;

	AssignIOComponents(inp_comps, input_tasks);		//create input tasks
	AssignIOComponents(out_comps, output_tasks);	//create output tasks

	FillInHoles(connecting_tasks);					//fill in the holes of the connecting tasks

	GenerateConnections(param.max_output_degree, param.min_output_degree, param.max_msg_size, param.min_msg_size);
}

void GenerateProcs(int num_procs, float min_bound, float max_bound)
{
	double bound;
	char proc_name[20];
	int num_procs_already = rand_sys.Get_Number_Of_Procs();

	//need to generate a network
	Network* net = new Network;
	NetworkProperty* net_prop = new NetworkProperty;
	double mbps = 1;			//hard code network speed to be 1 Mb / second
	net_prop->Set_Name(CString("Global Network"));
	net_prop->Set_Bit_Trans_Time((1.0*1000)/((float)mbps*1000000.0));
	net->Set_Property(net_prop);

	g_bus_speed = mbps;
	
	rand_sys.Add_Net(net);

	//generate an os, each proc will use this one os
	OS* os = new OS;
	OperatingSystemProperty* os_prop = new OperatingSystemProperty;
	os_prop->Set_Name(CString("Global OS"));
	os_prop->Set_Scheduling_Overhead(0);
	os_prop->Set_Context_Switch_Overhead(0);
	os_prop->Set_Timer_Overhead(0);
	os->Set_Property(os_prop);

	rand_sys.Add_OS(os);

	for(int i = 0; i < num_procs; i++)
	{
		Processor* proc = new Processor;
		ProcessorProperty* proc_prop = new ProcessorProperty;
		
		sprintf(proc_name, "Processor%i", i+num_procs_already);	//enumerate for proc name
		bound = Random(min_bound, max_bound);

		//add properties of processor
		proc_prop->Set_Bound(bound);
		proc_prop->Set_Name(CString(proc_name));
		proc_prop->Set_Type(0);
		proc_prop->Set_ID(i);

		//add property, os, and network to proc
		proc->Set_Property(proc_prop);
		proc->Set_OS(os);
		proc->Add_Network(net);

		//add the proc to the system and the network
		net->Add_Processor(proc);
		rand_sys.Add_Proc(proc);
	}
}

void GenerateComponents(int num_comps, float max_util, float min_util, double period, bool use_GME_hw)
{
	double total_util = GetTotalBound(use_GME_hw);	//get the sum of all bounds for the hardware
	double exec_time;								//that is being used
	double utilization = 0;
	double comp_util;
	char comp_name[20];
	int num_comps_already = rand_grp.Get_Number_Of_Nodes();

	int i = 0;

	//loop through until there are num_comps nodes
	while(rand_grp.Get_Number_Of_Nodes() < num_comps)
	{
		comp_util   =  Random(min_util, max_util);
		utilization += comp_util;					//keep track of the sum of all utilizations

		if(utilization > total_util*.95)			//stop loop if sum of all utilizations is
			break;									//above 95% of the sum of processor bounds

		exec_time = comp_util*period;
		sprintf(comp_name,"Coponent%i", num_comps_already + i++);		//enumerate for names

		TaskType* task = new TaskType;
		TaskProperty* task_prop = new TaskProperty;

		//set the tasks properties
		task_prop->Set_Exec_Time(exec_time);
		task_prop->Set_Name(CString(comp_name));
		task_prop->Set_Period(period);

		//add the properties to the task
		task->Set_Property(task_prop);

		//add the task to the graph
		rand_grp.Add_Node(task);

		//set the task to be a connecting task by default
		connecting_tasks[i] = task;
	}

}

void AssignIOComponents(int num_comps, map<int, TaskType*>& tasks)
{
	map<int, TaskType*>::iterator task_it;
	int last_index = connecting_tasks.rbegin()->first;	//the last index in the given set of tasks

	//remove num_comps of tasks from the set of conneting tasks and put them in the 
	//given set of tasks
	for(int i = 0; i < num_comps; ++i)
	{
		int index = Random(0, last_index);			//randomly choose an index

		task_it = connecting_tasks.find(index);
		if(task_it == connecting_tasks.end())		//position will be end if this index
		{											//was already removed - if so try again
			--i;
			continue;
		}

		tasks[i] = task_it->second;					//add task to set of tasks
		connecting_tasks.erase(task_it);			//remove task from connecting_tasks
	}
}

void FillInHoles(map<int, TaskType*>& tasks)
{	
	map<int, TaskType*> new_tasks;				//new set of tasks that will not have any holes
	map<int, TaskType*>::iterator task_it;

	int i = 0;
	for(task_it = tasks.begin(); task_it != tasks.end(); ++task_it)
		new_tasks[i++] = task_it->second;

	tasks = new_tasks;		//set tasks to the new set of tasks without any holes
}

void GenerateConnections(int max_output_degree, int min_output_degree, int max_msg_size, int min_msg_size)
{
	map<int, TaskType*> curr_list = input_tasks;	//the current list being iterated through (input_tasks or connecting_tasks)
	map<int, TaskType*>::iterator task_it;
	map<int, TaskType*>::iterator conn_task_it;

	int output_degree;			//how many output connections the task can have
	int msg_size;				//the size of the message in bytes

	bool update_index = false;	//update index once we are iterating through the connecting tasks
	int index = 0;				//next index after the task we are using in the connecting_tasks

	for(task_it = input_tasks.begin(); task_it != connecting_tasks.end(); ++task_it)
	{	
		if(task_it == input_tasks.end())	//finished input_tasks switch to connecting tasks
		{
			task_it = connecting_tasks.begin();
			curr_list = connecting_tasks;
			update_index = true;			//start updating the index
		}

		if(update_index)
			++index;

		TaskType* task = task_it->second;
		output_degree = Random(min_output_degree, max_output_degree);

		//generate each output connection
		for(int i = 0; i < output_degree; ++i)
		{
			double boundary = Random(double(0), double(1));	//used to randomly determine if we are connecting to a task from the output_tasks or from the connecting_tasks 
			int max_index, min_index;
			map<int, TaskType*>* tasks;

			//if the boundary is less than the ratio of output_tasks to output_tasks + 
			//available connecting_tasks use a task from output tasks to connect to the
			//given task, otherwise use connecting_tasks
			if(boundary <= double(output_tasks.size()) / (connecting_tasks.size() - index + output_tasks.size()))
			{
				max_index = output_tasks.size() - 1;
				min_index = 0;
				tasks     = &output_tasks;
			}
			else
			{
				max_index = connecting_tasks.size() - 1;
				min_index = index;
				tasks     = &connecting_tasks;
			}

			int conn_index = Random(min_index, max_index);
			int msg_size   = Random(min_msg_size, max_msg_size);

			conn_task_it = tasks->find(conn_index);

			if(conn_task_it == tasks->end())
				continue;

			TaskType* conn_task = conn_task_it->second;

			//if there is already a connection to this task don't add a new connection
			if(task->Get_Data_Connection_To(conn_task) != NULL)
				continue;

			//create new connection and its property
			ConnectionType* conn = new ConnectionType;
			ConnectionProperty* conn_prop = new ConnectionProperty;
			conn_prop->Set_Message_Size(msg_size);
			conn_prop->Set_Frequency(1.0/task->Property()->Get_Period());

			//add property and set to and from tasks for the connection
			conn->Set_Property(conn_prop);
			conn->Set_From(task);
			conn->Set_To(conn_task);

			//add the data connection to the task
			task->Add_Out_Data_Connection(conn);
			conn_task->Add_In_Data_Connection(conn);

			//add the connection to the graph
			rand_grp.Add_Edge(conn);
		}
	}
}

void StoreSystem(System& system, Graph& graph)
{
	CString szFileName;

	BOOL DoMyPromptFileName(CString& fileName, CString title, CString strFilter,int bOpenFileDialog);

	// get a file name from dialog.
	CString strFilter = "Text File (*.txt)";
	strFilter += (TCHAR)'\0';
	strFilter += _T("*.txt");
	strFilter += (TCHAR)'\0';

	if(!DoMyPromptFileName(szFileName, "File Save As", strFilter, FALSE))
		return;

	int pos = szFileName.Find(".txt");
	if(pos < 1)
		szFileName += ".txt";

	CFile f_save(szFileName, CFile::modeWrite | CFile::modeCreate);

	CString description = system.GetDescription() + graph.GetDescription();
	f_save.Write((LPCTSTR) description, strlen((LPCTSTR) description) );
	f_save.Close();
}

double GetTotalBound(bool use_GME_hw)
{
	list<Processor*>::iterator proc_it;
	double bound = 0;

	if(use_GME_hw)	//use the processors in GME_sys 
	{
		for(proc_it = GME_sys.Proc_Begin(); proc_it != GME_sys.Proc_End(); ++proc_it)
			bound += (*proc_it)->Property()->Get_Bound();
	}
	else			//use the processors in rand_sys
	{
		for(proc_it = rand_sys.Proc_Begin(); proc_it != rand_sys.Proc_End(); ++proc_it)
			bound += (*proc_it)->Property()->Get_Bound();
	}

	return bound;
}

void CleanUp()
{
	rand_sys.Clean();
	rand_grp.Clean();
}

void GenerateDescription(System* p_sys)
{
	CString description;
	CString line;

	list<Processor*>::iterator proc_it;
	double sum_of_bounds = 0;
	double min_bound = -1;
	double max_bound = -1;

	for(proc_it = p_sys->Proc_Begin(); proc_it != p_sys->Proc_End(); ++proc_it)
	{
		double bound = (*proc_it)->Property()->Get_Bound();

		if(min_bound == -1 || bound < min_bound)
			min_bound = bound;
		if(max_bound == -1 || bound > max_bound)
			max_bound = bound;

		sum_of_bounds += bound;
	}

	line.Format("\r\n********************* Hardware Summary *********************\r\n"); description+=line;
    line.Format("*\r\n*  Number of Processors:\t%d\r\n", p_sys->Get_Number_Of_Procs()); description+=line;
	line.Format("*  Average Processor Bound:\t%f\r\n", sum_of_bounds/p_sys->Get_Number_Of_Procs()); description+=line;
	line.Format("*  Minimum Processor Bound:\t%f\r\n", min_bound); description+=line;
	line.Format("*  Maximum Processor Bound:\t%f\r\n*", max_bound); description+=line;
	line.Format("\r\n************************************************************\r\n"); description+=line;

	line.Format("\r\n_____________________ Hardware Details _____________________\r\n\r\n"); description+=line;
	
	for(proc_it = p_sys->Proc_Begin(); proc_it != p_sys->Proc_End(); ++proc_it)
	{
		Processor* proc = *proc_it;
		line.Format("Name: %s\t\tBound: %f\r\n", proc->Property()->Get_Name(), proc->Property()->Get_Bound()); description+=line;
	}

	line.Format("\r\n____________________________________________________________\r\n"); description+=line;

	p_sys->SetDescription(description);
}

void GenerateDescription(Graph* p_graph)
{
	CString description;
	CString line;

	list<TaskType*>::iterator task_it;
	list<ConnectionType*>::iterator conn_it;

	double min_util = -1;
	double max_util = -1;
	double tot_util = 0;

	int min_msg_size = -1;
	int max_msg_size = -1;
	double tot_msg_size = 0;

	int min_conns = -1;
	int max_conns = -1;
	double tot_conns = 0;

	int input_comps = 0;
	int output_comps = 0;

	for(task_it = p_graph->Node_Begin(); task_it != p_graph->Node_End(); ++task_it)
	{
		TaskType* task = *task_it;

		double util = task->Property()->Get_Exec_Time() / task->Property()->Get_Period();
		int in_conns = task->Get_Number_Of_In_Data_Connection();
		int out_conns = task->Get_Number_Of_Out_Data_Connection();

		if(min_util == -1 || util < min_util)
			min_util = util;
		if(max_util == -1 || util > max_util)
			max_util = util;
		tot_util += util;

		if(min_conns == -1 || in_conns+out_conns < min_conns)
			min_conns = in_conns+out_conns;
		if(max_conns == -1 || in_conns+out_conns > max_conns)
			max_conns = in_conns+out_conns;
		tot_conns += in_conns+out_conns;

		if(in_conns == 0)
			++input_comps;
		if(out_conns == 0)
			++output_comps;
	}

	for(conn_it = p_graph->Edge_Begin(); conn_it != p_graph->Edge_End(); ++conn_it)
	{
		int msg_size = (*conn_it)->Property()->Get_Message_Size();

		if(min_msg_size == -1 || msg_size < min_msg_size)
			min_msg_size = msg_size;
		if(max_msg_size == -1 || msg_size > max_msg_size)
			max_msg_size = msg_size;
		tot_msg_size += msg_size;
	}

	line.Format("\r\n************* Component and Connection Summary *************\r\n"); description+=line;
    
	line.Format("*\r\n*  Number of Components:\t\t\t%d\r\n", p_graph->Get_Number_Of_Nodes()); description+=line;
	line.Format("*  Number of Input Components:\t\t%d\r\n", input_comps); description+=line;
	line.Format("*  Number of Output Components:\t\t%d\r\n", output_comps); description+=line;
	line.Format("*  Average Component Utilization:\t\t%f\r\n", tot_util/p_graph->Get_Number_Of_Nodes()); description+=line;
	line.Format("*  Minimum Component Utilization:\t\t%f\r\n", min_util); description+=line;
	line.Format("*  Maximum Component Utilization:\t\t%f\r\n", max_util); description+=line;

	line.Format("*\r\n*  Number of Connections:\t\t\t%d\r\n", p_graph->Get_Number_Of_Edges()); description+=line;
	line.Format("*  Average Connections per Component:\t\t%f\r\n", tot_conns/p_graph->Get_Number_Of_Nodes()); description+=line;
	line.Format("*  Minimum Connections per Component:\t\t%d\r\n", min_conns); description+=line;
	line.Format("*  Maximum Connections per Component:\t%d\r\n", max_conns); description+=line;
	line.Format("*  Average Message Size:\t\t\t%f\r\n", tot_msg_size/p_graph->Get_Number_Of_Edges()); description+=line;
	line.Format("*  Minimum Message Size:\t\t\t%d\r\n", min_msg_size); description+=line;
	line.Format("*  Maximum Message Size:\t\t\t%d\r\n*", max_msg_size); description+=line;
	
	line.Format("\r\n************************************************************\r\n"); description+=line;

	line.Format("\r\n_________________ Component and Connection _________________\r\n\r\n"); description+=line;
	
	for(task_it = p_graph->Node_Begin(); task_it != p_graph->Node_End(); ++task_it)
	{
		TaskType* task = *task_it;
		line.Format("Name: %s\t\tExec. Time: %f\tPeriod: %f\r\n", task->Property()->Get_Name(), task->Property()->Get_Exec_Time(), task->Property()->Get_Period()); description+=line;
		
		line.Format("In Connections:\r\n"); description+=line;
		for(conn_it = task->Get_In_Data_Edges_Begin(); conn_it != task->Get_In_Data_Edges_End(); ++conn_it)
		{
			TaskType* from_task = (*conn_it)->Get_From();
			int msg_size = (*conn_it)->Property()->Get_Message_Size();
			line.Format("\t%s\t%3d bytes\r\n", from_task->Property()->Get_Name(), msg_size); description+=line;
		}

		line.Format("Out Connections:\r\n"); description+=line;
		for(conn_it = task->Get_Out_Data_Edges_Begin(); conn_it != task->Get_Out_Data_Edges_End(); ++conn_it)
		{
			TaskType* to_task = (*conn_it)->Get_To();
			int msg_size = (*conn_it)->Property()->Get_Message_Size();
			line.Format("\t%s\t%3d bytes\r\n", to_task->Property()->Get_Name(), msg_size); description+=line;
		}
	}

	line.Format("\r\n____________________________________________________________\r\n"); description+=line;

	p_graph->SetDescription(description);
}