#ifndef COMM_MIN_H
#define COMM_MIN_H

#include <list>
#include <vector>
//#include <set>

//#include "load_balance.h"	//to obtain the function object processor_less_than

class ConnectionType;
class TaskType;
class Processor;
class CommMinEdge;

/* CommMinNode - a node used for the communication minimization algorithm.
   Has a list of tasks that make up the node and a list of CommMinEdges that
   connects the node to other nodes*/
class CommMinNode
{
public:
	CommMinNode() {}
	CommMinNode(TaskType* task);
	~CommMinNode();
	bool Merge(CommMinNode* node);

	bool operator == (CommMinNode n);			 

	inline void AddEdge(CommMinEdge* edge)		 {edges.push_back(edge);}
	void RemoveEdge(CommMinEdge* edge);
	inline std::list<CommMinEdge*>	GetEdges()	 {return edges;}

	inline std::list<CommMinEdge*>::iterator GetEdgeBegin() {return edges.begin();}
	inline std::list<CommMinEdge*>::iterator GetEdgeEnd()	{return edges.end();}

	inline std::list<TaskType*>::iterator GetTasksBegin()	{return tasks.begin();}
	inline std::list<TaskType*>::iterator GetTasksEnd()		{return tasks.end();}

	inline std::list<TaskType*>	GetTasks()					{return tasks;}
	inline void AssignProc(Processor* processor)			{proc = processor;}

	inline double GetTotalUtilization()						{return total_utilization;}

private:

	std::list<CommMinEdge*>	edges;
	std::list<TaskType*>	tasks;
	double					total_utilization;
	Processor*				proc;
};

/* CommMinEdge - an undirected edge used in the communication minimization algorithm.
   Has the two nodes that the edge connectes and the 'cost' of the edge which is the
   edges' bytes per second*/
class CommMinEdge
{
public:
	CommMinEdge(double bytes_per_second, CommMinNode* first_node, CommMinNode* second_node);

	inline void AddMessageSize(double bytes_per_second)	{bps += bytes_per_second;}
	inline double GetMessageSize()						{return bps;}
	inline void SetFirstConnectedNode(CommMinNode* node)	{node1 = node;}
	inline void SetSecondConnectedNode(CommMinNode* node)	{node2 = node;}
	inline CommMinNode* GetFirstConnectedNode()				{return node1;}
	inline CommMinNode* GetSecondConnectedNode()			{return node2;}

private:
	CommMinEdge();

	CommMinNode* node1;
	CommMinNode* node2;
	double bps;
};

/* CommMinGraph - a graph used for the communication minimization algorithm.  Has a 
   list of edges and nodes that makes up the graph*/
class CommMinGraph
{
public:
	inline std::list<CommMinNode*> GetNodes()	{return nodes;}
	inline std::list<CommMinEdge*> GetEdges()	{return edges;}

	inline void AddNode(CommMinNode* node)		{nodes.push_back(node);}
	void RemoveNode(CommMinNode* node);

	inline void AddEdge(CommMinEdge* edge)		{edges.push_back(edge);}
	void RemoveEdge(CommMinEdge* edge);

	CommMinNode* FindNodeFromTask(TaskType* task);	//find the node that a given task is in
	void		 RemoveEdgesFrom(CommMinNode* t);	//remove all the edges from the graph that are part of the given CommMinNode
private:
	std::list<CommMinNode*> nodes;
	std::list<CommMinEdge*> edges;
};

/* GraphPartition - a division of a CommMinGraph into groups.  Has a cost which is the
   total edge cost for the division of the two groups*/
struct GraphPartition
{
	GraphPartition() {}
	GraphPartition(const GraphPartition& g);
	GraphPartition& operator = (const GraphPartition& g);
	~GraphPartition() {}

	void AddTasksToFirst(CommMinNode* node);	//add all the tasks from the given node to the first group
	void AddTasksToSecond(CommMinNode* node);	//add all the tasks from the given node to the second group

	void AddTasks(int group_num, CommMinNode* node); //add all the tasks from the given node to the given group number 

	double GetFirstTotalUtilization();			//get the total utilization from all of the nodes in the first group
	double GetSecondTotalUtilization();			//get the total utilization from all of the nodes in the second group

	double GetTotalUtilization(int group_num);	//get the total utilization from all of the nodes in the given group number

	std::list<TaskType*> group1;
	std::list<TaskType*> group2;

	std::vector<std::list<TaskType*> >	group;
	double cost;
};

//map the tasks acording to the communication minimization algorithm
double Do_CommMin_Mapping();

/*
//get all of the tasks that are connected to task t
std::list<TaskType*> Get_Connected_Tasks(TaskType* t);

//get the sum of the utilizations of each task in the list
double Get_Total_Utilization(std::list<TaskType*> tasks);

//get the size of the sum of all connections to task t
double Get_Total_Connection_Size(TaskType* t);

//map tasks in the list to the processor given
void Put_Tasks_On_Proc(std::list<TaskType*> tasks, Processor* proc);

//check the list of tasks to see if any task has been assigned to a cpu
std::list<Processor*> Get_Procs_Assigned(std::list<TaskType*> tasks);

//remove the task that is not the root task that will cause the least amount 
//of additional overhead
std::list<TaskType*> Remove_Task(std::list<TaskType*> tasks, TaskType* root_task);*/

//run a modified min cut algorithm where the two groups must be below the given bounds
GraphPartition* MinCutAlgo();

//fill in the global graph variable with nodes and edges from the tasks and connections
//the model has 
void PopulateGraph();

//find the partition with the minimum cut for the current group of nodes
GraphPartition* MinimumCut();
void MinimumCut(GraphPartition* partition, std::list<CommMinNode*> nodes);
void MinimumCutRecursive(GraphPartition* partition, std::list<CommMinNode> nodes);

//Find the best node to add to the first group of the given partition from the
//group of nodes given 
CommMinNode* FindBestNode(GraphPartition partition, std::list<CommMinNode*> nodes);
CommMinNode* FindBestNodeNew(int group_num, GraphPartition* partition, std::list<CommMinNode*> nodes);

//Find the amount of edge cost that is saved by grouping the given node with the
//given partition
double find_saved_cost(CommMinNode* node, GraphPartition partition);
double findSavedCost(int group_num, CommMinNode* node, GraphPartition* partition);

//Find the total cost of the given partition for splitting the graph into the two 
//groups
double findPartitionCost(GraphPartition* partition);

//see if the given node is on the given list - return true if it is
bool onList(CommMinNode* node_check, std::list<CommMinNode*> checked_nodes);

//get the group number that the node is in in the given partition
int getGroup(CommMinNode* node, GraphPartition partition);
int getGroupNew(CommMinNode* node, GraphPartition* partition);

CommMinNode FindBestNodeRecurs(int group_num, GraphPartition* partition, std::list<CommMinNode> nodes);
bool valid_partition(GraphPartition* partition);

void show_partition(GraphPartition* partition);
/*
//function object to determine if the size of the data times the frequency for 
//connection c1 is greater then the size of the data times the frequency for 
//connection c2
class connection_greater_than
{
public:
	bool operator () (ConnectionType* c1, ConnectionType* c2);
};

//function object to determine if the total size of all connections of task t1
//is greater then the total size of all connections of task t2
class task_connection_size_greater_than
{
public:
	bool operator () (TaskType* t1, TaskType* t2);
};

//function object to determine if the total size of all connections of task t1
//is greater then the total size of all connections of task t2
class processor_most_connected_greater_than
{
public:
	processor_most_connected_greater_than(std::list<TaskType*> taskList, TaskType* rootTask) 
		{task_list = taskList; root_task = rootTask;}
	bool operator () (Processor* p1, Processor* p2);

private:
	processor_most_connected_greater_than();
	double greater_than_helper(Processor* p);

	std::list<TaskType*> task_list;
	TaskType* root_task;
};

typedef std::multiset<ConnectionType*, connection_greater_than> multisetConn;
typedef std::multiset<ConnectionType*, connection_greater_than>::iterator multisetConnIter;

typedef std::multiset<TaskType*, task_connection_size_greater_than> multisetTask;
typedef std::multiset<TaskType*, task_connection_size_greater_than>::iterator multisetTaskIter;

typedef std::multiset<Processor*, processor_less_than> multisetProc;
typedef std::multiset<Processor*, processor_less_than>::iterator multisetProcIter;

bool Assign_Group_To_Proc(std::list<TaskType*> tasks, multisetProc procs);*/

#endif