#include "comm_min-first_fit.h"

//global includes
#include <map>

//local includes
#include "debug.h"
#include "task_connection.h"
#include "graph.h"
#include "first_fit.h"

using namespace std;

double	abs(double d);	//absolute value for a double

//externals
extern Graph grp;
extern System sys;

//globals
bool optimal_comm_min;
__int64 num_steps;

//function objects used in heuristic version

//order edges by their bytes per second (Not realy message size)
struct CommMinEdgeGreaterThan
{
	bool operator () (CommMinEdge* c1, CommMinEdge* c2)
	{
		return c1->GetMessageSize() > c2->GetMessageSize();
	}
};

//order nodes by their utilization
struct CommMinNodeGreaterThan
{
	bool operator () (CommMinNode* n1, CommMinNode* n2)
	{
		return n1->GetTotalUtilization() > n2->GetTotalUtilization();
	}
};

//globals
CommMinGraph global_graph;

/** Static Helper Functions **/

//_______________________________________________________________________________________
// There are two different versions of this algorithm - one is optimal but is exponential
// so it is not currently used.  The other is O(PEN) where P is the number of processors,
// N is the number of components (nodes) and E is the number of connections (edges) and
// this is the algorithm that is currently implemented.  If the original algorithm is
// desired then comment out the specified lines in the Do_CommMinFirstFit_Mapping function
// These functions are used for the heuristic algorithm

//run the communication minimization algorithm with second order first fit heuristic
//algorithm, returns the amount of non-allocated utilization if the mapping fails
static double CommMinFirstFitHeuristic();

//returns true if the nodes in the global graph can be mapped to processors using the
//first fit algorithm - does not actually map them to the procs.
static bool FirstFitTest();

//run the first fit algorithm on the nodes in the global graph - does assign processors
//to nodes, returns the amount of non-allocated utilization if the mapping fails
static double FirstFitCommMin();

//sort all of the nodes such that they are in order of their utilization with highest
//utilization first
static void OrderNodes(multiset<CommMinNode*, CommMinNodeGreaterThan>& nodes);

//sort all of the processors such that 1. all processors that have any allocated tasks
//are before empty processors and 2. within these two groups they are ordered with highest
//bounds first
static void OrderProcs(multiset<Processor*, temp_processor_greater_than>& procs);

//assign the actual tasks to the processor that has been assigned to the node that the
//tasks are part of
static void AssignProcs();

//_________________________________________________________________________________________


/*** All of the functions below are for the algorithm not currently used. ***/


//---- These functions and varialbles are used for the (true) optimal algorithm ----

static map<int, TaskType*> g_tasks;		//all of the tasks enumerated
static map<int, Processor*> g_procs;	//all of the procs enumerated

double g_best_cost;

//run the optimal communication minimization algorithm through a depth first search - 
//at each level a task is assigned each of the processors, if some tasks do not have
//processors assigned to them and the utilization bound for each processor is not violated
//the expansion will continue, if all tasks are assigned to a processor the cost of this
//assignment is determined and stored if it is the best assignment so far
static bool OptimalCommMin();

//poulate the g_tasks and g_procs variables with all of the tasks and processors
static void PopulateGlobals();

//expand the node given by the given task number
static void Expand(int task_num);

//find the communication cost by the current assignments
static double FindComCost();

//store all of the task to processor assignments in the best_proc_assignment variable
static void StoreAssignments();

//add the tasks to the actual procs that is given in the best_proc_assignment variable
static void AssignOptimalProcs();
//-----------------------------------------------------------------------------

//--- These functions are not used at all ---

//get the number of nodes from the list of nodes that does not have a processor allocated to it
static int GetNumNodesNoProcs(list<CommMinNode*>& nodes);

//get the first node from the list of nodes that does not have a processor allocated to it
static CommMinNode* GetNodeNoProc(list<CommMinNode*>& nodes);

//remove tasks from the partition that already have a procesor allocated to them
static void RemoveTasksWithProcs(GraphPartition* partition);	

//remove all of the nodes from the given group number from the given partition that do not
//have a processor assigned to them
static void ClearTasksWithNoProcs(int group_num, GraphPartition* partition);	


/*************************************************************************************/
/*************** Communication Minimization with second order First Fit **************/
/*************************************************************************************/

double Do_CommMinFirstFit_Mapping()
{
	GraphPartition best_partition;
	double non_allocated_utility = 0;

	//reset all processors
	for(list<Processor *>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc_ptr = *proc_it;

		proc_ptr->Clear_Tasks();							//clear all tasks on the processor
		proc_ptr->Property()->Reset_Available_Space();		//give the processor its entire amount of space available
	}

	PopulateOriginalGraph(&global_graph);

	//___________________________________________________________________
	//Comment out the code within this block to run the original algorithm
	if(!optimal_comm_min)
	{
		non_allocated_utility = CommMinFirstFitHeuristic();
		if(non_allocated_utility)				//not all tasks could be mapped to a proc
			return non_allocated_utility;

		print_CommMin_Mapping(FindCommCost());	//print the results
		return 0;
	}

	if(!OptimalCommMin())	//run the optimal communication minimization algorithm
		return 1;			//error

	print_CommMin_Mapping(g_best_cost);	//print the results
	return 0;
	//___________________________________________________________________

	best_partition.SetGroupSize(sys.Get_Number_Of_Procs());	//set the numver of groups to the number of processors
	best_partition.cost = -1;								//set partition to be invalid
	best_partition.num_populated = 0;

	//run the min cut algorithm if all the tasks cannot be mapped to one processor
	//and there are at least 2 processors
	if(!AllTasksOnOneProc(&best_partition) && best_partition.group.size() >= 2)
		MinCutAlgo(&global_graph, &best_partition);
	else
	{
		best_partition.cost = findPartitionCost(&best_partition, &global_graph);
		RemoveTasksWithProcs(&best_partition);
	}

	if(best_partition.cost == -1)							//is the partition valid
	{
		non_allocated_utility = FirstFit();					//try one more time using the first fit algorithm - to get the nonallocated utility 
		if(non_allocated_utility)
			return non_allocated_utility;

		form_partition(&best_partition);					//first fit algorithm worked - form the partition
		best_partition.cost = findPartitionCost(&best_partition, &global_graph);
	}
	else
		map_tasks_to_procs(&best_partition);				//assign the tasks to the processors in the sys variable

//	print_CommMin_Mapping(best_partition.cost);				//print out the results
	print_CommMin_Mapping(FindCommCost());
	return 0;
}

double CommMinFirstFitHeuristic()
{
	multiset<CommMinEdge*, CommMinEdgeGreaterThan> edges;
	multiset<CommMinEdge*, CommMinEdgeGreaterThan>::iterator edge_it;

	list<CommMinEdge*>::iterator edge_list_it;

	CommMinNode* first;			//first node connected to an edge
	CommMinNode* second;		//second node connected to an edge
	CommMinNode* node;	

	Processor* proc_first;		//first node's processor
	Processor* proc_second;		//second node's processor
	Processor* proc;

	double non_allocated_util;	//utilization of tasks that could not be mapped to a proc.

	MergeNodesWithProcs(&global_graph);	//merge all of the nodes that are allocated to the same proc.

	num_steps = 0;

	//sort the edges in order of their communication cost
	for(edge_list_it = global_graph.Edges_Begin(); edge_list_it != global_graph.Edges_End(); ++edge_list_it)
		edges.insert(*edge_list_it);

	//go through each edge
	for(edge_it = edges.begin(); edge_it != edges.end(); ++edge_it)
	{
		++num_steps;

		first  = (*edge_it)->GetFirstConnectedNode();
		second = (*edge_it)->GetSecondConnectedNode();

		if(first == second)	//edges may be manipulated so that they are connected to the
			continue;		//same node, if this happens do not consider the edge

		proc_first  = first->GetAssignedProc();
		proc_second = second->GetAssignedProc();

		if(proc_first && proc_second)	//both nodes are allocated to a processor already,
			continue;					//there is nothing we can do
		else if((proc_first && !proc_second) || (!proc_first && proc_second))	//only 1 node has a proc
		{
			//set the proc to the non NULL proc and the node to the node with the NULL proc
			if(proc_first)
			{
				proc = proc_first;	
				node = second;
			}
			else
			{
				proc = proc_second;
				node = first;
			}

			//can this node be put onto the proc
			if(proc->Property()->Allocate_Space(node->GetTotalUtilization()))
			{
				first->AssignProc(proc);		//assign the proc to the first node

				first->QuickMerge(second);		//add utiliztion of second to first
				first->QuickEdgeMerge(second);	//add the edges of the second to the first
				first->QuickNodeMerge(second);	//add the tasks of the second to the first

				global_graph.RemoveNode(second);//remove the second node
				delete second;
			}
		}
		else
		{
			first->QuickMerge(second);			//add the utilization of the second to the first
			global_graph.RemoveNode(second);	//remove the second node from the graph

			if(FirstFitTest())					//can all of the tasks still be mapped to procs
			{									//using first fit
				first->QuickEdgeMerge(second);	//add the edges of the second to the first
				first->QuickNodeMerge(second);	//add the tasks of the second to the first

				delete second;
			}
			else								//merge no good - did not pass first fit test
			{
				first->UndoQuickMerge();		//undo the utilization merge
				global_graph.AddNode(second);	//add the node back to the graph
			} //else
		} //else
	} //for

	//run the first fit algorithm for all of the nodes that do not have a processor assigned
	//to them
	if(non_allocated_util = FirstFitCommMin())
		return non_allocated_util;

	//assign the tasks in each node to the processor assigned to the node
	AssignProcs();

	return 0;
}

bool FirstFitTest()
{
	multiset<CommMinNode*, CommMinNodeGreaterThan> nodes;
	multiset<CommMinNode*, CommMinNodeGreaterThan>::iterator node_it;
	
	multiset<Processor*, temp_processor_greater_than> procs;
	multiset<Processor*, temp_processor_greater_than>::iterator proc_it;

	double util;

	//order nodes in order of utilization and procs in order of their bounds
	OrderNodes(nodes);
	OrderProcs(procs);

	//map each node to a processor
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		util = (*node_it)->GetTotalUtilization();

		for(proc_it = procs.begin(); proc_it != procs.end(); ++proc_it)
		{
			++num_steps;

			//is there room for this node on this processor
			if((*proc_it)->Property()->Temp_Allocate_Space(util))
				break;
		}

		//could not find a processor to map the task to
		if(proc_it == procs.end())
			return false;
	}

	//all nodes found a proc - return true
	return true;
}

double FirstFitCommMin()
{
	multiset<CommMinNode*, CommMinNodeGreaterThan> nodes;
	multiset<CommMinNode*, CommMinNodeGreaterThan>::iterator node_it;
	
	multiset<Processor*, temp_processor_greater_than> procs;
	multiset<Processor*, temp_processor_greater_than>::iterator proc_it;

	double util;
	double non_allocated_util = 0;

	CommMinNode* node;

	//order nodes in order of utilization and procs in order of their bounds
	OrderNodes(nodes);
	OrderProcs(procs);

	//map each task to a processor
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		node = *node_it;
		util = node->GetTotalUtilization();

		for(proc_it = procs.begin(); proc_it != procs.end(); ++proc_it)
		{
			++num_steps;

			//is there room for this task on this processor
			if((*proc_it)->Property()->Temp_Allocate_Space(util))
			{
				node->AssignProc(*proc_it);	//assign the proc. to the node
				break;
			}
		}

		//could not find a processor to map the task to
		if(proc_it == procs.end())
			non_allocated_util += util;
	}

	return non_allocated_util;
}

void OrderNodes(multiset<CommMinNode*, CommMinNodeGreaterThan>& nodes)
{
	list<CommMinNode*>::iterator node_it;
	CommMinNode* node;

	//sorts in order of utilization
	for(node_it = global_graph.Nodes_Begin(); node_it != global_graph.Nodes_End(); ++node_it)
	{
		node = *node_it;

		++num_steps;

		if(!node->GetAssignedProc())	//only add node to the list if it is not assigned to a proc.
			nodes.insert(node);
	}
}

void OrderProcs(multiset<Processor*, temp_processor_greater_than>& procs)
{
	list<Processor*>::iterator proc_it;
	Processor* proc;

	//sorts in order of bounds
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		++num_steps;

		proc = *proc_it;
		proc->Property()->Set_Temp_Available_Space(proc->Property()->Get_Available_Space());
		procs.insert(proc);
	}
}

void AssignProcs()
{
	list<CommMinNode*>::iterator node_it;
	list<TaskType*>::iterator task_it;

	TaskType* task;
	Processor* proc;
	CommMinNode* node;

	//go through each node in the graph
	for(node_it = global_graph.Nodes_Begin(); node_it != global_graph.Nodes_End(); ++node_it)
	{
		node = *node_it;

		//get the proc (if any) assigned to the node
		if(proc = node->GetAssignedProc())
		{
			proc->Property()->Allocate_Space(node->GetTotalUtilization());

			//go through each task in the node and add it to the proc
			for(task_it = node->GetTasksBegin(); task_it != node->GetTasksEnd(); ++task_it)
			{
				task = *task_it;
				if(!task->Property()->Get_Alloc_Proc())
					proc->Add_Task(task);
			}
		}
	}
}

double FindCommCost()
{
	list<Processor*>::iterator proc_it1;
	list<Processor*>::iterator proc_it2;

	list<TaskType*>::iterator task_it1;
	list<TaskType*>::iterator task_it2;
	
	TaskType* task1;
	TaskType* task2;

	Processor* proc1;
	Processor* proc2;

	ConnectionType* conn1;
	ConnectionType* conn2;

	double cost = 0;

	//go through each proc.
	for(proc_it1 = sys.Proc_Begin(); proc_it1 != sys.Proc_End(); ++proc_it1)
	{
		proc1 = *proc_it1;

		//go through each task on the proc 1st proc
		for(task_it1 = proc1->Tasks_Begin(); task_it1 != proc1->Tasks_End(); ++task_it1)
		{
			task1 = *task_it1;

			//go through each proc after the position of the first proc
			for(proc_it2 = proc_it1; proc_it2 != sys.Proc_End(); ++proc_it2)
			{
				if(proc_it2 == proc_it1)	//2nd proc position starts at 1st so move past it
					continue;

				proc2 = *proc_it2;

				//go through each task on the 2nd proc
				for(task_it2 = proc2->Tasks_Begin(); task_it2 != proc2->Tasks_End(); ++task_it2)
				{
					task2 = *task_it2;

					//see if task1 and task2 are connected
					conn1 = task1->Get_Data_Connection_To(task2);
					conn2 = task2->Get_Data_Connection_To(task1);

					//if either connections are found add in the communication cost
					if(conn1)
						cost += conn1->Property()->Get_Message_Size() * conn1->Property()->Get_Frequency();
					if(conn2)
						cost += conn2->Property()->Get_Message_Size() * conn2->Property()->Get_Frequency();
				}
			}
		}
	}

	return cost;
}

double FindTotalComm()
{
	double total_comm = 0;
	for(list<TaskType*>::iterator task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
		for(list<ConnectionType*>::iterator conn_it = (*task_it)->Get_Out_Data_Edges_Begin(); conn_it != (*task_it)->Get_Out_Data_Edges_End(); ++conn_it)
			total_comm += (*conn_it)->Property()->Get_Message_Size() * (*conn_it)->Property()->Get_Frequency();
	return total_comm;
}

//_______________________________________________________________________________________
/*																						 */
/*********** All function definitions below are for the exponential algorithm ************/
/*_______________________________________________________________________________________*/

void MinCutAlgo(CommMinGraph* graph, GraphPartition* best_partition)
{
//	CommMinGraph rgraph = CommMinGraph(graph);				//copy the graph so original stays intact
	int groups_used = best_partition->num_populated;//GetNumGroupsUsed(best_partition);
	int groups_avaliable = best_partition->group.size() - groups_used;

	GraphPartition last_partition = *best_partition;

	MinimumCut(best_partition, graph);

	show_partition(best_partition);

	if(best_partition->cost == 0)	//if a partition of 0 is found return - best partition possible
		return;

	//find the best partition
	while(graph->GetNodes().size() >= groups_avaliable && graph->GetNodes().size() > 1)
	{
		if(GetNumNodesNoProcs(graph->GetNodes()) < 2)
			return;

		MinimumCut(&last_partition, graph);

		show_partition(&last_partition);

		//if the last partition is valid and either the best partition so far is invalid
		//or the cost of the las partition is less then the cost of the best partition
		//stor the last partition as the best partition
		if(last_partition.cost != -1 && (best_partition->cost == -1 || last_partition.cost < best_partition->cost))
			*best_partition = last_partition;	

		if(best_partition->cost == 0)
			return;

		//clear out groups that were temperarily used to find the partition
		for(int i = groups_used; i < last_partition.group.size(); ++i)
			ClearTasksWithNoProcs(i, &last_partition);
		//	last_partition.group[i].clear();
	}

	return;
}

void MinimumCut(GraphPartition* partition, CommMinGraph* graph)
{
	CommMinNode* s = NULL;
	CommMinNode* t = NULL;
	double cost;
	int group_num = partition->num_populated + 1;

	//find the group that needs to be populated
/*	for(group_num = 0; group_num < partition->group.size(); ++group_num)
		if(partition->group[group_num].empty())
		{
			group_num++;
			break;
		}*/

	list<CommMinNode*> nodes = graph->GetNodes();	//copy the nodes so the list can be
													//be modified without changing the
													//actual graph

	s = nodes.front();
	if(!s->GetAssignedProc())
		partition->AddTasks(group_num, s);
	nodes.pop_front();

	while(!nodes.empty())
	{
		if(nodes.size() == 1 || GetNumNodesNoProcs(nodes) == 1)
		{
			if(nodes.size() == 1)
				t = nodes.front();
			else
				t = GetNodeNoProc(nodes);
			if(!t->GetAssignedProc())
				partition->AddTasks(group_num - 1, t);
			nodes.pop_front();
			break;
		}

		s = FindBestNode(group_num, partition, nodes, graph);	//find the best node to add to the partition
		if(!s->GetAssignedProc())
			partition->AddTasks(group_num, s);
		nodes.remove(s);
	}

	if(valid_partition(partition))								//is partition valid
		partition->cost = findPartitionCost(partition, graph);
//	else if(partition->group[partition->group.size()-1].size())
	else if(group_num == partition->group.size() - 1)
		partition->cost = -1;									//the partition does not have any free groups
	else
	{
		CommMinGraph graph_copy = CommMinGraph(graph);

//		list<attachment> attachments = StoreAttachments(t);		//the partition does have free groups
																//split up the graph more	
		CommMinNode* new_t = graph_copy.FindNodeFromTask(t->GetTasks().front());

		//remove the last node added (t) from the copied graph
		graph_copy.RemoveNode(new_t);
		graph_copy.RemoveEdgesFrom(new_t);

		partition->num_populated++;

		partition->group[group_num].clear();	//clear out the tasks that were in the strongest connected group
		MinCutAlgo(&graph_copy, partition);		//run min cut algorithm on modified graph

		partition->num_populated--;
	}


	//s and t should always be non NULL - this if statement is just an added precaution
	if(s && t)
	{
		//merge s and t together and remove t and its edges from the graph
		s->Merge(t);
		graph->RemoveNode(t);
		graph->RemoveEdgesFrom(t);

		delete t;
	}

} //MinimumCut

void PopulateOriginalGraph(CommMinGraph* graph)
{
	graph->clean();		//clear out any existing stuff

	//add in all of the nodes - needs to be done before adding edges
	for(list<TaskType*>::iterator task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		TaskType* task_ptr = *task_it;
		Processor* proc;

		++num_steps;

		if(proc = task_ptr->Property()->Get_Alloc_Proc())
		{
			double utilization = task_ptr->Property()->Get_Exec_Time() / task_ptr->Property()->Get_Period();
			if(proc->Property()->Allocate_Space(utilization))
				proc->Add_Task(task_ptr);
		}

		CommMinNode* new_task = new CommMinNode(task_ptr);
		new_task->AssignProc(proc);
		graph->AddNode(new_task);
	}

	//add in all of the edges
	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		//add in connections that are coming out of this connection
		for(list<ConnectionType*>::iterator conn_it = (*task_it)->Get_Out_Data_Edges_Begin(); 
		conn_it != (*task_it)->Get_Out_Data_Edges_End(); ++conn_it)
		{
			TaskType* from = (*conn_it)->Get_From();
			TaskType* to = (*conn_it)->Get_To();

			++num_steps;

			//get the node that the tasks are part of 
			CommMinNode* node1 = graph->FindNodeFromTask(from);
			CommMinNode* node2 = graph->FindNodeFromTask(to);

			double bps = (*conn_it)->Property()->Get_Frequency() * (*conn_it)->Property()->Get_Message_Size();

			CommMinEdge* new_edge = new CommMinEdge(bps, node1, node2);

			//add the new edge to its nodes and to the graph
			node1->AddEdge(new_edge);
			node2->AddEdge(new_edge);
			graph->AddEdge(new_edge);
		}
	}
}

CommMinNode* FindBestNode(int group_num, GraphPartition* partition, list<CommMinNode*> nodes, CommMinGraph* graph)
{
	CommMinNode* best;
	double saved_cost;
	double best_saved_cost;

	best			= nodes.front();	//initialize the best node as the first node
	best_saved_cost	= findSavedCost(group_num, best, partition, graph); //find the saved cost of adding it to the partition

	for(list<CommMinNode*>::iterator it = nodes.begin(); it != nodes.end(); ++it)
	{
		//find the saved cost of adding it to the partition
		saved_cost = findSavedCost(group_num, *it, partition, graph);

		//check to see if this node saves more then the previous best node
		if(saved_cost > best_saved_cost)
		{
			best_saved_cost = saved_cost;
			best = *it;
		}
	}

	return best;
}

double findSavedCost(int group_num, CommMinNode* node, GraphPartition* partition, CommMinGraph* graph)
{
	double saved_cost = 0;
	list<TaskType*> tasks = partition->group[group_num];
	list<CommMinNode*> checked_nodes;

	for(list<TaskType*>::iterator it = tasks.begin(); it != tasks.end(); ++it)
	{
		TaskType* task = *it;
		CommMinNode* it_node = graph->FindNodeFromTask(task);

		//check if this node has already been tested, if so don't check it again
		if(onList(it_node, checked_nodes))
			continue;
		checked_nodes.push_back(it_node);

		list<CommMinEdge*> edges = it_node->GetEdges();

		//go through each of the edges in the partition and see if they are connected 
		//to the given node - if so add in the edge's bps to the saved cost
		for(list<CommMinEdge*>::iterator conn_it = it_node->GetEdgeBegin(); conn_it != it_node->GetEdgeEnd(); ++conn_it)
		{
			CommMinEdge* edge = *conn_it;

			CommMinNode* first_node = edge->GetFirstConnectedNode();
			CommMinNode* second_node = edge->GetSecondConnectedNode();

			if(first_node == node || second_node == node)
				saved_cost += edge->GetMessageSize();
		}
	}

	return saved_cost;
}

double findPartitionCost(GraphPartition* partition, CommMinGraph* graph)
{
	double cost = 0;
	double edge_cost;

	CommMinNode* node1;
	CommMinNode* node2;

	list<CommMinEdge*> edges = global_graph.GetEdges();

	//go through each edge and determine if its nodes are in different groups - if so
	//then add the edge cost to the total cost
	for(list<CommMinEdge*>::iterator edge_it = edges.begin(); edge_it != edges.end(); ++edge_it)
	{
		int node1group;
		int node2group;

		node1		= (*edge_it)->GetFirstConnectedNode();
		node2		= (*edge_it)->GetSecondConnectedNode();
		edge_cost   = (*edge_it)->GetMessageSize();

		node1group  = getGroup(node1, partition, &global_graph);
		node2group  = getGroup(node2, partition, &global_graph);

		if(node1group != node2group)
			cost += edge_cost;
	}

	return cost;
}

bool onList(CommMinNode* node_check, list<CommMinNode*>  checked_nodes)
{
	for(list<CommMinNode*>::iterator it = checked_nodes.begin(); it != checked_nodes.end(); ++it)
		if(node_check == *it) return true;

	return false;
}

int getGroup(CommMinNode* node, GraphPartition* partition, CommMinGraph* graph)
{
	list<CommMinNode*> checked_nodes;

	for(int i = 0; i < partition->group.size(); ++i)
	{
		list<TaskType*> tasks = partition->group[i];
		for(list<TaskType*>::iterator task_it = tasks.begin(); task_it != tasks.end(); ++task_it)
		{
			CommMinNode* node_check = graph->FindNodeFromTask(*task_it);

			//make sure that the node has not already been previously checked
			if(onList(node_check, checked_nodes))
				continue;
			else
				checked_nodes.push_back(node_check);
			
			//found the node
			if(node_check == node)
				return i;
		}
	}

	//could not find the node
	return -1;
}


void show_partition(GraphPartition* partition)
{
	string name;
	int group;

	//go through each group and each task in the group
	for(group = 0; group < partition->group.size(); ++group)
		for(list<TaskType*>::iterator it = partition->group[group].begin(); it != partition->group[group].end(); ++it)
			name = (*it)->Property()->Get_Name();
}

bool valid_partition(GraphPartition* partition)
{
	multiset<list<TaskType*>, partition_group_greater_than> tasks;
	multiset<Processor*, processor_greater_than> procs;

	if(partition->group.size() == 0)	//if there are no groups then the partition cannot be valid
		return false;

	//sort the groups in the partition by their total utilization
	for(int i = 0; i < partition->group.size(); ++i)
		tasks.insert(partition->group[i]);

	//sort the processors by their available space
	for(list<Processor*>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
		procs.insert(*proc_it);

	//check to see if the largest group can be mapped to the processor to the greatest
	//bound, the second largest to the second greatest, and so on
	multiset<list<TaskType*>, partition_group_greater_than>::iterator task_it = tasks.begin();
	for(multiset<Processor*, processor_greater_than>::iterator it = procs.begin();
	it != procs.end(); ++it)
	{
		if(GetTotalUtilization(*task_it) > (*it)->Property()->Get_Bound())
			return false;

		MapGroupToProc(*task_it, *it, partition);
		task_it++;
	}

	return true;
}

bool AllTasksOnOneProc(GraphPartition* partition)
{
	map<Processor*, int> processor_to_index;
	list<TaskType*> all_tasks;
	multiset<Processor*, processor_greater_than> procs;
	int i = 0;

	if(partition->group.size() == 0)	//no procs so cannot put them all on one processor
		return false;

	//sort the processors by their available space
	for(list<Processor*>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
		procs.insert(*proc_it);

	//set up mapping from the processor to the index of where it is
	for(multiset<Processor*, processor_greater_than>::iterator ms_proc_it = procs.begin(); ms_proc_it != procs.end(); ++ms_proc_it)
	{
		partition->procs[i] = *ms_proc_it;
		processor_to_index[*ms_proc_it] = i++;
	}

	//get all of the tasks
	for(list<TaskType*>::iterator all_it = grp.Node_Begin(); all_it != grp.Node_End(); ++all_it)
	{
		TaskType* task = *all_it;

		if(!task->Property()->Get_Alloc_Proc())
			all_tasks.push_back(task);
		else
		{
			int group_index = processor_to_index[task->Property()->Get_Alloc_Proc()];
			partition->group[group_index].push_back(task);
		}
	}

	//can all of the tasks be put on the processory with the largest bound
	if(GetTotalUtilization(all_tasks) < (*procs.begin())->Property()->Get_Bound())
	{
		partition->group[0].insert(partition->group[0].end(), all_tasks.begin(), all_tasks.end());
		partition->procs[0] = *procs.begin();
		partition->cost = 0;
		return true;
	}

	return false;
}

list<attachment> StoreAttachments(CommMinNode* node)
{
	attachment a;
	list<attachment> attachments;

	//go through each edge of the node
	for(list<CommMinEdge*>::iterator edge_it = node->GetEdgeBegin(); edge_it != node->GetEdgeEnd(); ++edge_it)
	{
		double bps = (*edge_it)->GetMessageSize();
		CommMinNode* other_node;

		//find the other node that is part of the edge
		if((*edge_it)->GetFirstConnectedNode() == node)
			other_node = (*edge_it)->GetSecondConnectedNode();
		else
			other_node = (*edge_it)->GetFirstConnectedNode();

		//get all of the tasks that are part of the other node
		for(list<TaskType*>::iterator task_it = other_node->GetTasksBegin(); task_it != other_node->GetTasksEnd(); ++task_it)
		{
			a.bytes_per_second = bps;
			a.task = *task_it;
			attachments.push_back(a);
		}
	}

	return attachments;
}

bool partition_group_greater_than::operator () (list<TaskType*> t1, list<TaskType*> t2)
{
	return GetTotalUtilization(t1) > GetTotalUtilization(t2);
}

double GetTotalUtilization(list<TaskType*> tasks)
{
	double u = 0;

	for(list<TaskType*>::iterator it = tasks.begin(); it != tasks.end(); ++it)
		u += (*it)->Property()->Get_Exec_Time() / (*it)->Property()->Get_Period();

	return u;
}

int GetNumGroupsUsed(GraphPartition* partition)
{
	for(int i = 0; i < partition->group.size(); ++i)
		if(partition->group[i].empty()) break;

	return i;
}

void map_tasks_to_procs(GraphPartition* partition)
{
	for(int i = 0; i < partition->group.size(); ++i)
	{
		for(list<TaskType*>::iterator task_it = partition->group[i].begin(); task_it != partition->group[i].end(); ++task_it)
		{
			if((*task_it)->Property()->Get_Alloc_Proc())
				continue;

			double utilization = (*task_it)->Property()->Get_Exec_Time() / (*task_it)->Property()->Get_Period();
			partition->procs[i]->Property()->Allocate_Space(utilization);	//this function should always return true - there is a bug otherwise
			partition->procs[i]->Add_Task(*task_it);
		}
	}
}

void form_partition(GraphPartition* partition)
{
	int i = 0;
	for(list<Processor*>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		for(list<TaskType*>::iterator task_it = (*proc_it)->Tasks_Begin(); task_it != (*proc_it)->Tasks_End(); ++task_it)
			partition->group[i].push_back(*task_it);
		++i;
	}
}

void MapGroupToProc(list<TaskType*> task_list, Processor* proc, GraphPartition* partition)
{
	//find the group that the list represents
	for(int i = 0; i < partition->group.size(); ++i)
		if(task_list.front() == partition->group[i].front()) break;

	//if the group was found set the processor to it
	if(i < partition->group.size())
		partition->procs[i] = proc;
}

int GetNumNodesNoProcs(list<CommMinNode*>& nodes)
{
	int rval = 0;
	list<CommMinNode*>::iterator node_it;

	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
		if((*node_it)->GetAssignedProc() == NULL) ++rval;
	
	return rval;
}

CommMinNode* GetNodeNoProc(list<CommMinNode*>& nodes)
{
	list<CommMinNode*>::iterator node_it;

	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
		if((*node_it)->GetAssignedProc() == NULL) return *node_it;

	return NULL;
}

void RemoveTasksWithProcs(GraphPartition* partition)
{
	list<TaskType*>::iterator task_it;

	for(int i = 0; i < partition->group.size(); ++i)
	{
		task_it = partition->group[i].begin();
		while(task_it != partition->group[i].end())
		{
			TaskType* task = *task_it;
			if(task->Property()->Get_Alloc_Proc())
				task_it = partition->group[i].erase(task_it);
			else
				++task_it;
		}
	}
}

void ClearTasksWithNoProcs(int group_num, GraphPartition* partition)
{
	list<TaskType*>::iterator task_it;

	for(task_it = partition->group[group_num].begin(); task_it !=partition->group[group_num].end();)
	{
		TaskType* task = *task_it;

		if(task->Property()->Get_Alloc_Proc())
		{
			++task_it;
			continue;
		}
		
		task_it = partition->group[group_num].erase(task_it);
	}
}

bool OptimalCommMin()
{
	PopulateGlobals();
	g_best_cost = -1;				//set there to be no best_proc_assignment 

	num_steps = 0;					//reset the number of steps taken

	Expand(0);						//expand the first node

	if(g_best_cost == -1)			//cost == -1 if there were no legal assignemts for
		return false;				//every task

	AssignOptimalProcs();			//make the assignments permanent

	return true;
}

void PopulateGlobals()
{
	list<TaskType*>::iterator task_it;
	list<Processor*>::iterator proc_it;

	g_tasks.clear();
	g_procs.clear();

	int i = 0;
	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
		g_tasks[i++] = *task_it;
	
	i=0;
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		g_procs[i++] = *proc_it;
		(*proc_it)->Property()->Set_Temp_Available_Space(0);	//reset the available space
	}
}

void Expand(int task_num)
{
	TaskType* task = g_tasks[task_num];
	double util = task->Property()->Get_Exec_Time() / task->Property()->Get_Period();	//utilization

	//try each processor for the current task
	for(int proc_num = 0; proc_num != sys.Get_Number_Of_Procs(); ++proc_num)
	{
		++num_steps;		//increment number of steps taken

		if(g_best_cost == 0)	//if a cost of 0 is found no need to keep searching
			return;

		Processor* proc = g_procs[proc_num];

		double old_util = proc->Property()->Get_Temp_Available_Space();
		double new_util = old_util + util;

		//can this task be assigned to this proc without violating its bound
		if(new_util > proc->Property()->Get_Bound())
			continue;

		proc->Property()->Set_Temp_Available_Space(new_util);	//updated utilziation
		task->Property()->SetProcNum(proc_num);					//set the processor number

		if(task_num == grp.Get_Number_Of_Nodes() - 1)	//is this the last task
		{
			double cost = FindComCost();

			//is this the best assignment so far or is this the first assignment 
			if(cost < g_best_cost || g_best_cost == -1)
			{
				g_best_cost = cost;
				StoreAssignments();
			}
		}
		else
			Expand(task_num+1);		//move on to the next task

		//reset to previous value before assignment
		proc->Property()->Set_Temp_Available_Space(old_util);
	}
}

double FindComCost()
{
	list<TaskType*>::iterator task_it;
	list<ConnectionType*>::iterator conn_it;
	double cost = 0;

	//go through each task
	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		TaskType* t1 = *task_it;

		//go through each out data connection
		for(conn_it = t1->Get_Out_Data_Edges_Begin(); conn_it != t1->Get_Out_Data_Edges_End(); ++conn_it)
		{
			++num_steps;

			ConnectionType* conn = *conn_it;

			TaskType* t2 = conn->Get_To();

			//if the processors are not the same add in the communication cost
			if(t1->Property()->GetProcNum() != t2->Property()->GetProcNum())
				cost += conn->Property()->Get_Message_Size() * conn->Property()->Get_Frequency();
		}
	}

	return cost;
}

void StoreAssignments()
{
	list<TaskType*>::iterator task_it;

	//store the current assignments in best_proc_assignment 
	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		TaskType* task = *task_it;
		task->Property()->SetBestProcNum(task->Property()->GetProcNum());
	}
}

void AssignOptimalProcs()
{
	list<TaskType*>::iterator task_it;

	//go through each task
	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		TaskType* task = *task_it;

		//get the proc_num for this task from the best assignment
		int proc_num = task->Property()->GetBestProcNum();
		Processor* proc = g_procs[proc_num];	//get the actual processor

		double util = task->Property()->Get_Exec_Time() / task->Property()->Get_Period();

		proc->Property()->Allocate_Space(util);	//allocate the space
		proc->Add_Task(task);					//add the task
	}
}
