#include "comm_min-load_balance.h"

//local includes
#include "comm_min-first_fit.h"
#include "debug.h"
#include "task_connection.h"
#include "graph.h"

//global includes
#include <list>
#include <set>
#include <map>

using namespace std;

/******************************************************************************************/
/*************** Communication Minimization with second order Load Balancing **************/
/******************************************************************************************/

/*** Static Helper Classes ***/

//struct to keep track a possible valid merge between two nodes and the extra utilization
//that the merge would cause processors to go over their ideal bound
struct PossibleMerge
{
	double extra_util;
	CommMinNode* node1;
	CommMinNode* node2;
};

//list of tasks and their total utilization
struct TaskList
{
	list<TaskType*> tasks;
	double utilization;
	Processor* proc;
};

//struct to keep track of a processor, its ideal_bound (the bound that the processor would
//prefer to stay below but is less then or equal to its true bound), how much utilization
//can be added until the ideal bound is reached, and a node that may be assigned to it 
struct ProcWrapper
{
	ProcWrapper(Processor* in_proc = NULL, double in_bound = 0): proc(in_proc), 
		ideal_bound(in_bound), ideal_space(in_bound) {}

	double ideal_bound;
	double ideal_space;
	Processor* proc;
	TaskList task_list;
};

//returns the sum of the cost of all of the node's edges 
static double	FindTotalEdgeCost(CommMinNode* node);

/* Function Objecst used for sorting*/

//a processor wrapper is greater than another if its ideal space is highest
class ProcWrpGreaterThan
{
public:
	bool operator()(ProcWrapper* p1, ProcWrapper* p2)
	{
		return p1->ideal_space > p2->ideal_space;
	}
};

//a task list is greater than another if its total utilization is highest
class TaskListGreaterThan
{
public:
	bool operator()(TaskList list1, TaskList list2)
	{
		return list1.utilization > list2.utilization;
	}
};

//a node is greater than another if its total edge cost is highest
class NodeEdgeCostGreaterThan
{
public:
	bool operator()(CommMinNode* node1, CommMinNode* node2)
	{
		return FindTotalEdgeCost(node1) > FindTotalEdgeCost(node2);
	}
};

//an edge is greater than another if its communication cost is highest
class EdgeCostGreaterThan
{
public:
	bool operator()(CommMinEdge* edge1, CommMinEdge* edge2)
	{
		return edge1->GetMessageSize() > edge2->GetMessageSize();
	}
};

/*** Static Helper Functions ***/

//run the first algorithm - Merge the first valid pair of nodes where one node has the greatest total edge 
//cost and the node other node is the one that is connected to the first node with the greatest edge cost
static double	DoAlgo1(multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps);

//run the second algorithm - Merge the first valid pair of nodes where the nodes are the two connected by
//the edge with the highest communication cost
static double	DoAlgo2(multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps);

//returns the utilization of all of the nodes in the global_graph
static double	FindTotalUtilization();

//populates the given list and determines the ideal bound for each member in the list - returns the amount
//of utilization that could not be put onto a processor, 0 if successful
static double	DetermineProcCommMinBounds(list<ProcWrapper*>& wrappers);

//check to see if merging node1 and node2 is valid - if so merge the two nodes, returns true if nodes were
//merged and false if they were not 
static bool		CheckAndMerge(CommMinNode* node1, CommMinNode* node2, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps);

//returns the position of the TaskList in the global variable task_lists
static multiset<TaskList, TaskListGreaterThan>::iterator	GetTaskList(CommMinNode*);

//merge the two nodes and update the global variable task_lists
static void		MergeAndUpdate(CommMinNode* node1, CommMinNode* node2);

//assign processors to the tasks that are in global variable task_lists
static void		AssignProcs(multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps);

//find the best pair of unconnected nodes to merge - returns true if a valid pair of nodes
//is found, this pair is found from the multiset nodes variable and put into the merged_nodes 
//variable, the checked variable allows the function to only checked nodes that have not
//already been checked and the proc_wrps variable is the set of the processor wrappers that are used
static bool		MergeUnconnected(PossibleMerge& merged_nodes, multiset<CommMinNode*, NodeEdgeCostGreaterThan> nodes, multimap<CommMinNode*, CommMinNode*> checked, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps);

//merge all of the nodes in the list of nodes given that have processors assigned to them,
//then put all of these nodes onto the associated processor wrapper 
static void		MergeNodesWithProcs(list<CommMinNode*>& nodes, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps);

//find the processor wrapper that is associated with the given processor from the set of
//processor wrappers given - NULL is returned if this isn't found
static ProcWrapper* FindWrapper(Processor* proc, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps);

//check to see if merging the two given TaskLists onto the processor given in the wrapper variable
//is valid, a -1 is returned if it is valid, otherwise the utilization over the processors ideal bound
//is returned, so if 0 is returned then the merge will keep the total utilization under the processors
//ideal bound. The list of the processor wrappers is also needed.
static double	CheckMergeOntoProc(ProcWrapper* wrapper, TaskList& list1, TaskList& list2, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps);

//Find the nth highest list that does not have a processor assigned to it from global task_lists function 
//that is not the list that is in the not_this_list variable based upon TaskListGreaterThan function object
static void		FindNthHighestList(int n, TaskList& task_list, TaskList& not_this_list);

//Find the nth highest processor from the proc_wrps variable that does not already have tasks assigned to it
static Processor* FindNthHighestAvailableProc(int n, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps);

//debug functions to view the tasks in each node and the tasks in each list
static void		ShowNodes();

static void		ShowLists();


//externals
extern Graph grp;
extern System sys;
extern CommMinGraph global_graph;
extern __int64 num_steps;

//static globals
static PossibleMerge best_valid_merge;						//the best valid merge of two nodes that was found
static multiset<TaskList, TaskListGreaterThan> task_lists;	//list of a list of tasks that keeps its elements in order 
															//by the total utilization of all of the tasks in the list

//global variables to choose which algorithm to run
bool algo1;
bool algo2;	
												
double Do_CommMinLoadBalance_Mapping()
{
	double rval;
	multiset<ProcWrapper*, ProcWrpGreaterThan> ms_proc_wrps;
	list<ProcWrapper*> proc_wrps;
	list<ProcWrapper*>::iterator wrp_it;

	list<CommMinNode*> nodes;
	list<CommMinNode*>::iterator node_it;

	//get rid of any old data
	global_graph.clean();
	task_lists.clear();

	num_steps = 0;

	PopulateOriginalGraph(&global_graph);

	//reset all processors
	for(list<Processor *>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc_ptr = *proc_it;

		++num_steps;

		proc_ptr->Clear_Tasks();							//clear all tasks on the processor
		proc_ptr->Property()->Reset_Available_Space();		//give the processor its entire amount of space available
	}

	if((rval = DetermineProcCommMinBounds(proc_wrps)) != 0)
		return rval;	//not all tasks would be able to be mapped to a processor

	//insert processor wrappers into ms_proc_wrps to keep them in order of their ideal bound
	for(wrp_it = proc_wrps.begin(); wrp_it != proc_wrps.end(); ++wrp_it)
	{
		++num_steps;
		ms_proc_wrps.insert(*wrp_it);
	}

	nodes = global_graph.GetNodes();
	
	//merge all of the nodes that have the same processor assigned to it
	MergeNodesWithProcs(nodes, ms_proc_wrps);
	
	TaskList t_list;
	//insert the nodes' lists into the task lists to keep them in order of their total utilization
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		++num_steps;
		t_list.tasks = (*node_it)->GetTasks();
		t_list.utilization = (*node_it)->GetTotalUtilization();
		t_list.proc = (*node_it)->GetAssignedProc();
		task_lists.insert(t_list);
	}

	//choose which algorithm to run
	if(algo1)
	{
		if(rval = DoAlgo1(ms_proc_wrps))
			return rval;
	}
	else if(algo2)
	{
		if(rval = DoAlgo2(ms_proc_wrps))
			return rval;
	}
	else
		return -1;	//algo1 or algo2 should always be selected

	GraphPartition partition;
	partition.SetGroupSize(sys.Get_Number_Of_Procs());

	AssignProcs(ms_proc_wrps);	//assign procs to each of the tasks that are in the nodes in global_graph
	form_partition(&partition);	//form a partition from the processor assignments
	partition.cost = findPartitionCost(&partition, &global_graph);	//find the cost of this partition

	//print the results
	print_CommMin_Mapping(partition.cost);

	return 0;
}

double DoAlgo1(multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps)
{
	multimap<CommMinNode*, CommMinNode*> checked;				//variable to store if a two nodes have already been checked to see if they can be merged
	multimap<CommMinNode*, CommMinNode*>::iterator checked_it; 

	list<CommMinNode*>::iterator node_it;
	list<CommMinNode*> nodes = global_graph.GetNodes();

	list<CommMinEdge*>::iterator edge_it;
	list<CommMinEdge*> edges;

	multiset<CommMinNode*, NodeEdgeCostGreaterThan> ms_nodes;	//set holds the list of nodes in order of their total edge cost
	multiset<CommMinNode*, NodeEdgeCostGreaterThan>::iterator ms_node_it;
	multiset<CommMinNode*, NodeEdgeCostGreaterThan>::iterator ms_node_it1;
	multiset<CommMinNode*, NodeEdgeCostGreaterThan>::iterator ms_node_it2;

	multiset<CommMinEdge*, EdgeCostGreaterThan> ms_edges;
	multiset<CommMinEdge*, EdgeCostGreaterThan>::iterator ms_edge_it;

	//sort nodes by the sum of the cost of each nodes' edges and resize checked
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		++num_steps;
		ms_nodes.insert(*node_it);
	}

	//go through and merge nodes until there are the same number of nodes as processors
	while(global_graph.GetNodes().size() > proc_wrps.size())
	{	
		//reset best_valid_merge
		best_valid_merge.node1 = NULL;
		best_valid_merge.node2 = NULL;
		best_valid_merge.extra_util = 0;

		//reset checked data
		checked.clear();

		//go through each node starting with the node with the greatest total edge cost
		for(ms_node_it = ms_nodes.begin(); ms_node_it != ms_nodes.end(); ++ms_node_it)
		{
			edges = (*ms_node_it)->GetEdges();
			ms_edges.clear();

			//sort edges by their edge cost
			for(edge_it = edges.begin(); edge_it != edges.end(); ++edge_it)
			{
				++num_steps;
				ms_edges.insert(*edge_it);
			}

			//go through each edge that is connected to this node starting with the edge with the highest cost
			for(ms_edge_it = ms_edges.begin(); ms_edge_it != ms_edges.end(); ++ms_edge_it)
			{
				CommMinNode* node1 = *ms_node_it;
				CommMinNode* node2 = (*ms_edge_it)->GetOtherNode(*ms_node_it);

				//make sure this pair of nodes was not already checked
				for(checked_it = checked.lower_bound(node1); checked_it != checked.upper_bound(node1); ++checked_it)
				{
					++num_steps;
					if(checked_it->second == node2) break;
				}
					
				if(checked_it != checked.upper_bound(node1))
					continue;

				//mark the pair of nodes as checked
				checked.insert(make_pair(node1, node2));
				checked.insert(make_pair(node2, node1));

				//check to see if these nodes can be merged - if so merge them
				if(CheckAndMerge(node1, node2, proc_wrps))
				{
					//update ms_nodes to remove merged nodes and to keep nodes in order

					//Note: could not get ms_nodes.find(node2) to work so used this search loop
					for(ms_node_it2 = ms_nodes.begin(); ms_node_it2 != ms_nodes.end(); ++ms_node_it2)
					{
						++num_steps;
						if(*ms_node_it2 == node2)
						{
							ms_nodes.erase(ms_node_it2);
							break;
						}
					}

					ms_nodes.erase(ms_node_it);	//remove node and reinsert it to keep nodes in order
					ms_nodes.insert(node1);
					break;
				}
			}

			if(ms_edge_it != ms_edges.end())	//nodes were merged
				break;
		}

		if(ms_node_it == ms_nodes.end())		//could not find a pair of nodes that could be merged to keep 
		{										//the total utilization of processors below their ideal bounds
			CommMinNode* node1;
			CommMinNode* node2;
			double extra_util;

			//a valid merge was found that caused some processor to be above its ideal bound
			if(best_valid_merge.node1)
			{
				node1 = best_valid_merge.node1;
				node2 = best_valid_merge.node2;
				extra_util = best_valid_merge.extra_util;
			}
			else	//no best valid merge was found - try checking unconnected components
			{
				PossibleMerge merged_nodes;
				if(MergeUnconnected(merged_nodes, ms_nodes, checked, proc_wrps))
				{
					//a valid merge was found
					node1 = merged_nodes.node1;
					node2 = merged_nodes.node2;
					extra_util = merged_nodes.extra_util;
				}
				else
					return -1;		//no solution was found
			}

			if(extra_util != -1)	//this should never be true - just an added precaution
				MergeAndUpdate(node1, node2);

			//Note: could not get ms_nodes.find(node2) to work so used this search loop
			ms_node_it1 = ms_nodes.end();
			ms_node_it2 = ms_nodes.end();

			for(ms_node_it = ms_nodes.begin(); ms_node_it != ms_nodes.end(); ++ms_node_it)
			{
				++num_steps;

				if(*ms_node_it == node1)
				{
					ms_node_it1 = ms_node_it;
					if(ms_node_it2 != ms_nodes.end())
						break;
				}
				if(*ms_node_it == node2)
				{
					ms_node_it2 = ms_node_it;
					if(ms_node_it1 != ms_nodes.end())
						break;
				}
			}

			ms_nodes.erase(ms_node_it1);	//remove and reinsert node1 to keep nodes in order
			ms_nodes.erase(ms_node_it2);
			ms_nodes.insert(node1);
		}
	}

	return 0;
}

double DoAlgo2(multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps)
{
	multimap<CommMinNode*, CommMinNode*> checked;			//variable to store if a two nodes have already been checked to see if they can be merged
	multimap<CommMinNode*, CommMinNode*>::iterator checked_it; 

	multiset<CommMinEdge*, EdgeCostGreaterThan> ms_edges;
	multiset<CommMinEdge*, EdgeCostGreaterThan>::iterator ms_edge_it;

	list<CommMinNode*> nodes;
	list<CommMinNode*>::iterator node_it;
	multiset<CommMinNode*, NodeEdgeCostGreaterThan> ms_nodes;

	list<CommMinEdge*>::iterator edge_it;
	list<CommMinEdge*> edges;

	//go through and merge nodes until there are the same number of nodes as processors
	while(global_graph.GetNodes().size() > proc_wrps.size())
	{
		//reset best_valid_merge
		best_valid_merge.node1 = NULL;
		best_valid_merge.node2 = NULL;
		best_valid_merge.extra_util = 0;

		//clear checked nodes map
		checked.clear();

		edges = global_graph.GetEdges();
		ms_edges.clear();

		//sort edges by their cost
		for(edge_it = edges.begin(); edge_it != edges.end(); ++edge_it)
		{
			++num_steps;
			ms_edges.insert(*edge_it);
		}

		//go through each edge starting with the one with the highest cost
		for(ms_edge_it = ms_edges.begin(); ms_edge_it != ms_edges.end(); ++ms_edge_it)
		{
			CommMinNode* node1 = (*ms_edge_it)->GetFirstConnectedNode();
			CommMinNode* node2 = (*ms_edge_it)->GetSecondConnectedNode();

			//check to see if this pair of nodes has already been checked
			for(checked_it = checked.lower_bound(node1); checked_it != checked.upper_bound(node1); ++checked_it)
			{
				++num_steps;
				if(checked_it->second == node2) break;
			}
				
			if(checked_it != checked.upper_bound(node1))
				continue;

			//mark pair of nodes as checked
			checked.insert(make_pair(node1, node2));
			checked.insert(make_pair(node2, node1));

			//see if the two nodes that are part of the edge can be merged - if so merge them
			if(CheckAndMerge(node1, node2, proc_wrps))
				break;
		}

		//
		if(ms_edge_it == ms_edges.end())	//could not find a pair of nodes that could be merged to keep 
		{									//the total utilization of processors below their ideal bounds
			if(best_valid_merge.node1)	
				MergeAndUpdate(best_valid_merge.node1, best_valid_merge.node2);
			else
			{
				//a pair of connected nodes could not be found to merge that would keep a processor
				//below its bounds - try unconnected nodes

				PossibleMerge merged_nodes;
				nodes = global_graph.GetNodes();

				//sort nodes in order of their total edge cost
				ms_nodes.clear();
				for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
				{
					++num_steps;
					ms_nodes.insert(*node_it);
				}

				//try to merged unconnected nodes
				if(MergeUnconnected(merged_nodes, ms_nodes, checked, proc_wrps))
				{
					if(merged_nodes.extra_util != -1)	//this statement should always be true
						MergeAndUpdate(merged_nodes.node1, merged_nodes.node2);
				}
				else
					return -1;	//-1 signifies that the algorithm could not be completed
			}
		}
	}

	return 0;
}

double FindTotalEdgeCost(CommMinNode* node)
{
	double cost = 0;
	list<CommMinEdge*>::iterator edge_it;

	for(edge_it = node->GetEdgeBegin(); edge_it != node->GetEdgeEnd(); ++edge_it)
		cost += (*edge_it)->GetMessageSize();

	return cost;
}

double FindTotalUtilization()
{
	list<CommMinNode*>::iterator node_it;
	list<CommMinNode*> nodes = global_graph.GetNodes();
	double utilization = 0;

	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
		utilization += (*node_it)->GetTotalUtilization();

	return utilization;
}

double DetermineProcCommMinBounds(list<ProcWrapper*>& wrappers)
{
	list<ProcWrapper*> empty;
	ProcWrapper* curr_wrapper;
	list<ProcWrapper*>::iterator wrapper_it;
	list<Processor*>::iterator proc_it;
	int non_full_procs = sys.Get_Number_Of_Procs();

	double total_utilization = FindTotalUtilization();	//get the total utilization of all components
	double util_avg;

	//populate wrappers with a wrapper around each processor
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
		wrappers.push_back(new ProcWrapper(*proc_it));

	//go until there is no more utiliation to distribute
	while(total_utilization && total_utilization > .000001) //second condition takes care of errors when using doubles 
	{
		bool util_changed = false;
		util_avg = total_utilization / non_full_procs;

		for(wrapper_it = wrappers.begin(); wrapper_it != wrappers.end(); ++wrapper_it)
		{
			double ideal_bound = (*wrapper_it)->ideal_bound;
			double true_bound = (*wrapper_it)->proc->Property()->Get_Bound();

			//see if the added util_avg will excede the processor's bound
			if(util_avg + ideal_bound <= true_bound)
			{
				(*wrapper_it)->ideal_bound += util_avg;
				(*wrapper_it)->ideal_space += util_avg;
				total_utilization -= util_avg;
				util_changed = true;
			}
			else if(abs(true_bound - ideal_bound) > .000001)	//if the processors ideal bound does not already equal
			{													//its true bound - errors for doubles accounted for
				double added_bound = true_bound - ideal_bound;	//find the amount the bound can hold
				(*wrapper_it)->ideal_bound += added_bound;
				(*wrapper_it)->ideal_space += added_bound;
				total_utilization -= added_bound;
				non_full_procs--;								//processor is full so there is one fewer non_full_proc
				util_changed = true;
			}
		}

		//if all of the processors are full or there was no utilization change
		if(!non_full_procs || !util_changed)
			return total_utilization;
	}

	return 0;
}

bool CheckAndMerge(CommMinNode* node1, CommMinNode* node2, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps)
{
	list<TaskType*>::iterator task_it;
	multiset<ProcWrapper*, ProcWrpGreaterThan>::iterator wrp_it;
	multiset<TaskList, TaskListGreaterThan>::iterator it;
	multiset<TaskList, TaskListGreaterThan>::iterator insert_it;
	multiset<TaskList, TaskListGreaterThan>::iterator it1;
	multiset<TaskList, TaskListGreaterThan>::iterator it2;
	Processor* proc1;
	Processor* proc2;
	ProcWrapper* wrapper = NULL;
	double extra_util = -1;
	
	//empty task list
	TaskList empty;
	empty.proc = NULL;
	empty.utilization = 0;

	proc1 = node1->GetAssignedProc();
	proc2 = node2->GetAssignedProc();

	//both nodes were already assigned different processors - cannot merge
	if(proc1 && proc2 && proc1 != proc2)
		return false;

	if(!proc1)			//proc1 has the assigned processor to use (if any)
		proc1 = proc2;

	it1 = GetTaskList(node1);				//get the position of the two lists in task_lists
	it2 = GetTaskList(node2);

	TaskList list1 = *it1;					//get the actual lists
	TaskList list2 = *it2;
	TaskList copy_list1 = list1;			//store original version of list1

	task_lists.erase(it1);					//remove lists from task_lists
	task_lists.erase(it2);

	//merge two lists together
	list1.utilization += list2.utilization;
	list1.tasks.insert(list1.tasks.end(), list2.tasks.begin(), list2.tasks.end());
	list1.proc = proc1;

	insert_it = task_lists.insert(list1);	//reinsert list1

	if(proc1)								//a processor was assigned for this list
	{
		wrapper = FindWrapper(proc1, proc_wrps);
		extra_util = CheckMergeOntoProc(wrapper, list1, empty, proc_wrps);
	}
	else									//find a processor for this task
	{	//go through each processor
		for(wrp_it = proc_wrps.begin(); wrp_it != proc_wrps.end(); ++wrp_it)
		{
			double curr_extra_util;

			++num_steps;

			//see if putting the merged node and any other tasks on this processor would be valid
			curr_extra_util = CheckMergeOntoProc(*wrp_it, list1, (*wrp_it)->task_list, proc_wrps);
	
			if(curr_extra_util == -1)	//invalid
				continue;
			else if(!curr_extra_util)	//valid - under ideal bound
			{
				extra_util = curr_extra_util;
				break;
			}
			else if(extra_util == -1 || curr_extra_util < extra_util)	//valid - but above ideal bound check to see
			{															//if the utilization above the ideal bound is 
				wrapper = *wrp_it;										//less then the previous best one
				extra_util = curr_extra_util;
			}
		}
	}

	if(!extra_util)
	{
		//everything is ok merge node1 and node2
		node1->Merge(node2);
		node1->AssignProc(proc1);
		global_graph.RemoveNode(node2);
		global_graph.RemoveEdgesFrom(node2);
		return true;
	}
	else if(extra_util == -1)
	{
		//merge is not valid
		task_lists.erase(insert_it);
		task_lists.insert(copy_list1);
		task_lists.insert(list2);
		return false;
	}
	else
	{
		//if the extra utilization is smaller then the previous best_valid_merge or 
		//there was no previous best_valid_merge
		if(best_valid_merge.node1 == NULL || extra_util < best_valid_merge.extra_util)
		{
			best_valid_merge.extra_util = extra_util;
			best_valid_merge.node1 = node1;
			best_valid_merge.node2 = node2;
		}

		task_lists.erase(insert_it);
		task_lists.insert(copy_list1);
		task_lists.insert(list2);
		return false;
	}
}

double abs(double d)
{
	if(d < 0)
		return -d;
	return d;
}

multiset<TaskList, TaskListGreaterThan>::iterator GetTaskList(CommMinNode* node)
{
	multiset<TaskList, TaskListGreaterThan>::iterator list_it;
	TaskType* task;

	//go through each task_list in the list
	for(list_it = task_lists.begin(); list_it != task_lists.end(); ++list_it)
		for(list<TaskType*>::iterator it = list_it->tasks.begin(); it != list_it->tasks.end(); ++it)
		{
			task = *it;
			if(task == *(node->GetTasksBegin())) //if task is equal to the first task in the node
				return list_it;
		}

	return task_lists.end();
}

void MergeAndUpdate(CommMinNode* node1, CommMinNode* node2)
{
	multiset<TaskList, TaskListGreaterThan>::iterator it;
	multiset<TaskList, TaskListGreaterThan>::iterator it1;
	multiset<TaskList, TaskListGreaterThan>::iterator it2;
	Processor* proc = node1->GetAssignedProc();

	//if node1 did not have a processor assigned to it try node2
	if(!proc)
		proc = node2->GetAssignedProc();

	it1 = GetTaskList(node1);
	it2 = GetTaskList(node2);

	TaskList list1 = *it1;
	TaskList list2 = *it2;

	task_lists.erase(it1);
	task_lists.erase(it2);

	//merge the two lists
	list1.utilization += list2.utilization;
	list1.tasks.insert(list1.tasks.end(), list2.tasks.begin(), list2.tasks.end());
	list1.proc = proc;

	task_lists.insert(list1);	//reinsert list1 to keep in order

	//merge actual nodes
	node1->Merge(node2);
	node1->AssignProc(proc);
	global_graph.RemoveNode(node2);
	global_graph.RemoveEdgesFrom(node2);
}

void AssignProcs(multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps)
{
	multiset<TaskList, TaskListGreaterThan>::iterator list_it;
	list<TaskType*>::iterator task_it;
	Processor* proc;
	int n = 1;

	//go through each task list
	for(list_it = task_lists.begin(); list_it != task_lists.end(); ++list_it)
	{
		if(list_it->proc)			//if the task was assigned a processor use its processor
			proc = list_it->proc;
		else						//task was not assigned a processor use the Nth highest available processor and increment n 
			proc = FindNthHighestAvailableProc(n++, proc_wrps);

		//assign each task on the list to the processor
		for(task_it = list_it->tasks.begin(); task_it != list_it->tasks.end(); ++task_it)
		{
			++num_steps;

			double util = (*task_it)->Property()->Get_Exec_Time() / (*task_it)->Property()->Get_Period();
			proc->Property()->Allocate_Space(util);
			proc->Add_Task(*task_it);

			(*task_it)->Property()->Set_Proc(proc);
		}
	}
}


bool MergeUnconnected(PossibleMerge& merged_nodes, multiset<CommMinNode*, NodeEdgeCostGreaterThan> nodes, multimap<CommMinNode*, CommMinNode*> checked, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps)
{
	multiset<CommMinNode*, NodeEdgeCostGreaterThan>::reverse_iterator node_it1;
	multiset<CommMinNode*, NodeEdgeCostGreaterThan>::reverse_iterator node_it2;
	multimap<CommMinNode*, CommMinNode*>::iterator checked_it;

	//reset best_valid_merge
	best_valid_merge.node1 = NULL;
	best_valid_merge.node2 = NULL;
	best_valid_merge.extra_util = 0;

	CommMinNode* node1;
	CommMinNode* node2;

	//go through the nodes starting from the node with the smallest total edge cost
	for(node_it1 = nodes.rbegin(); node_it1 != nodes.rend(); ++node_it1)
	{
		//go through the nodes starting from the node with the smallest total edge cost just after node_it1
		for(node_it2 = node_it1; node_it2 != nodes.rend(); ++node_it2)
		{
			if(node_it1 == node_it2)	//move to node after node_it1
				continue;

			node1 = *node_it1;
			node2 = *node_it2;

			++num_steps;

			//see if this pair of nodes was already checked
			for(checked_it = checked.lower_bound(node1); checked_it != checked.upper_bound(node1); ++checked_it)
			{
				++num_steps;
				if(checked_it->second == node2) break;
			}
					
			if(checked_it != checked.upper_bound(node1))
				continue;

			//mark this pair of nodes as checked
			checked.insert(make_pair(node1, node2));
			checked.insert(make_pair(node2, node1));

			//try to merge these two nodes
			if(CheckAndMerge(node1, node2, proc_wrps))
			{
				merged_nodes.node1 = node1;
				merged_nodes.node2 = node2;
				merged_nodes.extra_util = -1;
				return true;
			}
		}
	}

	//no two nodes were found to merge that would keep the processor below its ideal bound, see if there
	//were two that would keep the processor below its true bound
	if(best_valid_merge.node1)
	{
		merged_nodes.node1 = best_valid_merge.node1;
		merged_nodes.node2 = best_valid_merge.node2;
		merged_nodes.extra_util = best_valid_merge.extra_util;
		return true;
	}

	//no nodes could be merged together without violating a processors true bound
	return false;
}

void MergeNodesWithProcs(list<CommMinNode*>& nodes, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps)
{
	list<CommMinNode*>::iterator node_it;
	list<Processor*>::iterator proc_it;
	CommMinNode* merged_node;

	//go through each processor
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc = *proc_it;
		merged_node = NULL;

		//go through each node
		node_it = nodes.begin();
		while(node_it != nodes.end())
		{
			CommMinNode* curr_node = *node_it;
			if(curr_node->GetAssignedProc() == proc)	//is the assigned processor for this node equal to 
			{											//the current processor
				if(merged_node)		//was there already a node found with this processor
				{
					//merge the two nodes
					merged_node->Merge(curr_node);		
					global_graph.RemoveNode(curr_node);
					global_graph.RemoveEdgesFrom(curr_node);
					node_it = nodes.erase(node_it);		//erase this node from the list and get the next iterator position
				}
				else
				{
					merged_node = curr_node;	//this is the first node that this processor was assigned to
					++node_it;					//this node will merge all the other nodes that have this processor
				}								//assigned to it
			}
			else
				++node_it;
		}

		if(merged_node)	//there was at least one node that had this processor assigned to it - register it with its wrapper
		{
			ProcWrapper* proc_wrp = FindWrapper(proc, proc_wrps);
			proc_wrp->task_list.tasks = merged_node->GetTasks();
			proc_wrp->task_list.utilization = merged_node->GetTotalUtilization();
			proc_wrp->task_list.proc = proc;
		}
		else			//no nodes were assigned to this processor - set associated wrapper to be empty
		{
			ProcWrapper* proc_wrp = FindWrapper(proc, proc_wrps);
			proc_wrp->task_list.utilization = 0;
			proc_wrp->task_list.proc = NULL;
		}
	}
}

ProcWrapper* FindWrapper(Processor* proc, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps)
{
	multiset<ProcWrapper*, ProcWrpGreaterThan>::iterator wrp_it;

	for(wrp_it = proc_wrps.begin(); wrp_it != proc_wrps.end(); ++wrp_it)
		if((*wrp_it)->proc == proc) return *wrp_it;

	return NULL;
}

double CheckMergeOntoProc(ProcWrapper* wrapper, TaskList& list1, TaskList& list2, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps)
{
	int i = 1;
	double extra_util = 0;
	multiset<ProcWrapper*, ProcWrpGreaterThan>::iterator wrp_it;

	//if list1 or list2 is assigned a processor it must be the same as the processor in the wrapper variable
	if(list1.proc && list1.proc != wrapper->proc)
		return -1;

	if(list2.proc && list2.proc != wrapper->proc)
		return -1;

	double merged_util = list1.utilization + list2.utilization;

	//go through each processor wrapper
	for(wrp_it = proc_wrps.begin(); wrp_it != proc_wrps.end(); ++wrp_it)
	{
		double utilization;

		//if this processor wrapper is equal to the one we are trying to put the two task lists onto
		if(*wrp_it == wrapper)
			utilization	= merged_util;
		else
		{
			if((*wrp_it)->task_list.tasks.size())	//does this processor already have tasks on it - if so use this list
				utilization	= (*wrp_it)->task_list.utilization;
			else
			{
				TaskList t_list;
				FindNthHighestList(i++, t_list, list1);	//find the Nth highest task list that does not have a processor assigned to it
				if(t_list.tasks.size())					//if a task list was found use its utilization
					utilization = t_list.utilization;
				else									//no task list was found so the utilization of the processor is 0
					utilization = 0;					
			}
		}

		double ideal_bound = (*wrp_it)->ideal_bound;
		double true_bound  = (*wrp_it)->proc->Property()->Get_Bound();

		if(utilization*.99999 > ideal_bound)			//if the utilization is greater than desired
		{
			if(utilization > true_bound)				//if the utilization is greater than possible 
				return -1;
			
			extra_util += utilization - ideal_bound;	//utilization that could excedes the ideal bounds
		}
	}

	return extra_util;
}

void FindNthHighestList(int n, TaskList& task_list, TaskList& not_this_list)
{
	multiset<TaskList, TaskListGreaterThan>::iterator list_it;
	int current_n = 1;

	//go through each task list in order of their total utilizaition
	for(list_it = task_lists.begin(); list_it != task_lists.end(); ++list_it)
	{
		TaskList curr_list = *list_it;
		if(curr_list.tasks.size() && curr_list.tasks.front() == not_this_list.tasks.front())	//make sure the list isn't
			continue;																			//the not_this_list list

		if(curr_list.proc == NULL && current_n++ == n)	//if the current list is empty check to see if current_n equals 
		{												//the given n - if so set the task list to be this list
			task_list = curr_list;
			return;
		}
	}
}

Processor* FindNthHighestAvailableProc(int n, multiset<ProcWrapper*, ProcWrpGreaterThan>& proc_wrps)
{
	int current_n = 1;
	multiset<ProcWrapper*, ProcWrpGreaterThan>::iterator wrp_it;

	//go through each processor wrapper in order of their available ideal space
	for(wrp_it = proc_wrps.begin(); wrp_it != proc_wrps.end(); ++wrp_it)
	{
		ProcWrapper* proc_wrp = *wrp_it;
		if(proc_wrp->task_list.tasks.empty() && current_n++ == n)	//if the current processor does not have tasks
			return proc_wrp->proc;									//assigned to it and the current_n is equal to 
																	//n return the processor wrapper
	}

	return NULL;
}

/* Functions used just for debugging */

void ShowNodes()
{
	list<CommMinNode*> nodes = global_graph.GetNodes();
	CString name;
	int group = 0;

	for(list<CommMinNode*>::iterator node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		for(list<TaskType*>::iterator task_it = (*node_it)->GetTasksBegin(); task_it != (*node_it)->GetTasksEnd(); ++task_it)
			name = (*task_it)->Property()->Get_Name();
		group++;
	}
}

void ShowLists()
{
	multiset<TaskList, TaskListGreaterThan>::iterator list_it;
	list<TaskType*>::iterator task_it;

	int list_num = 0;
	CString task_name;
	CString proc_name;
	TaskList curr_list;

	for(list_it = task_lists.begin(); list_it != task_lists.end(); ++list_it)
	{
		list_num++;
		curr_list = *list_it;

		if(curr_list.proc)
			proc_name = curr_list.proc->Property()->Get_Name();
		else
			proc_name = "No Proc";

		for(task_it = curr_list.tasks.begin(); task_it != curr_list.tasks.end(); ++task_it)
			task_name = (*task_it)->Property()->Get_Name();
	}
}
