#ifndef COMM_MIN_H
#define COMM_MIN_H

/*
 * Communication Minimization Algorithm
 * ------------------------------------ 
 *
 * The communication minimization algorithm will map the tasks to the processors
 * with the least amount of communcation between processors as possible.  This is 
 * done using a modified version of maximum cardinality search algorithm described
 * in the article "A Simple Min-Cut Algorithm" by Mechthild Stoer and Frank Wagner
 * found in the Journal of the ACM, Vol. 44, No. 4, July 1997, pp. 585-591.  The
 * algorithm that was derived is a recursive version of the original algorithm that
 * also takes into consideration different numbers of processors (meaning differnt
 * numbers of groups) and the bounds of the processors.
 *
 * 1.  Find the minimum cut of the phase and group the tasks into two groups by 
 *     running the original algorithm.
 * 2.  Check to see if this produces a valid partition - meaning that the groups
 *     produced can be mapped to the processors without violating their bound conditions
 * 3a. If the partition is valid then find the edge cost that this produces and
 *     store it if it is better then best partition that was previously found and
 *     continue the original algorithm.
 * 3b. If the partition is not valid then check to see if the partition can be broken up
 *     into more groups - there can be as many groups as there are processors
 * 4.  If the partition can be broken into more tasks then take the group of nodes that 
 *     has the strongest connections (see the paper for details) and run this algorithm
 *     on that group.
 * 5.  This is continued until a partition is valid or if the partition cannot be broken
 *     into any more groups.  Each valid partition that is produced is checked against
 *     the best partition and is stored if its cost is cheaper then best partition
 *     found previously.
 *
 * NOTE: Compiling this code will cause many warnings if you are using MSV C++
 * since MSV C++ does not handle some of the standard containers and function
 * objects very well.
 */

#include <list>

#include "comm_min_graph.h"

//map the tasks acording to the 1st order communication minimization 2nd order first
//fit algorithm
double Do_CommMinFirstFit_Mapping();

//run the modified recursive min cut algorithm explained above on the given graph
//Assumptions - The size of the partitions groups is 2 or more, there are at least
//2 groups in the given partition that can be populated
void MinCutAlgo(CommMinGraph* graph, GraphPartition* partition);

//fill in the global graph variable with nodes and edges from the tasks and connections
//the model has 
void PopulateOriginalGraph(CommMinGraph* graph);

//find the partition with the minimum cut for the current group of nodes
//Assumptions - There are at least 2 nodes that can be merged in the given graph
void MinimumCut(GraphPartition* partition, CommMinGraph* graph);

//Find the best node to add to the given group of the given partition from the
//group of nodes given that is part of the given graph
CommMinNode* FindBestNode(int group_num, GraphPartition* partition, std::list<CommMinNode*> nodes, CommMinGraph* graph);

//Find the amount of edge cost that is saved by grouping the given node with the
//given partition
double findSavedCost(int group_num, CommMinNode* node, GraphPartition* partition, CommMinGraph* graph);

//Find the total cost of the given partition for splitting the graph into the two 
//groups
double findPartitionCost(GraphPartition* partition, CommMinGraph* graph);

//see if the given node is on the given list - return true if it is
bool onList(CommMinNode* node_check, std::list<CommMinNode*> checked_nodes);

//get the group number that the node is in in the given partition that is part of the
//given graph
int getGroup(CommMinNode* node, GraphPartition* partition, CommMinGraph* graph);

//Is the given partition valid - meaning that the groups in the partition can be mapped
//to the processors in the system and not violate their bounds'
//Assumptions - the group size of the partition is equal to the number of processors in
//the system
bool valid_partition(GraphPartition* partition);

//Cann all of the tasks be put on one processor - if so populate the given partition
bool AllTasksOnOneProc(GraphPartition* partition);

//Store the attachments of the given node into the list that is returned
std::list<attachment> StoreAttachments(CommMinNode* node);

//Find the total utilization of all of the tasks that are given
double GetTotalUtilization(std::list<TaskType*> tasks);

//Find the total number of groups that are used by the given partition
int GetNumGroupsUsed(GraphPartition* partition);

//map the tasks in the partition to their assigned processors in the system
//Assumptions - The group mapping to the procesors done by the given partition 
//              is valid 
void map_tasks_to_procs(GraphPartition* partition);

//fill in the given partition from mapping to the system variables that was done
//previously
//Assumptions - the group size of the partition is equal to the number of processors in
//the system
void form_partition(GraphPartition* partition);

//Find the group in the given partition that is equal to the given list of tasks and
//map the group to the given processor
void MapGroupToProc(std::list<TaskType*> task_list, Processor* proc, GraphPartition* partition);

//Run the communication minimization algorithm with second order load balance mapping.
//returns the amount of utilization that could not be mapped to the processor or -1 if
//there is an error but there may be a solution that the algorithm could not find
double Do_CommMinLoadBalance_Mapping();

//Function object used for sorting groups in a partition - tests to see if the total
//utilization of t1 is greater then the total utilization of t2
class partition_group_greater_than
{
public:
	bool operator () (std::list<TaskType*> t1, std::list<TaskType*> t2);
};

//Used soley for debugging - easily shows the cost of the partition, name, and group
//of each of its members in the debuging window
void show_partition(GraphPartition* partition);
#endif