#include "..\stdafx.h"
#include "windows.h"
#include "graphalgo.h"
#include "deadline.h"
#include "debug.h"

typedef struct
{
    int pathlen;        // Number of nodes in the path segment
    float distance;     // Total execution time in the path segment
    float deadline;     // Deadline of the path segment
    float offset;       // Offset of the path segment
} path_segment_data_dd;

static void get_path_segment_data (path_segment_data_dd&,TaskType *,
        list<sssp_entry>*);
static void allocate_slack(path_segment_data_dd&,TaskType *,
        deadline_distribution_policies);
static void set_deadlines_parents (TaskType*,path_segment_data_dd&,
        deadline_distribution_policies);

void Deadline_Distribute (Graph& g, deadline_distribution_policies ddp)
{
    TaskType *dttp;
    list<sssp_entry>* dsel=SSSP_Bellman_Ford(g);
    list<TaskType *>* rdttl=Topo_Sort(g);
    list<TaskType *>::iterator dti;
    list<TaskType*> dttl;


    for(dti=rdttl->begin();dti!=rdttl->end();dti++)
    {
        dttl.push_front(*dti);
    }

    for(dti=g.Node_Begin();dti!=g.Node_End();dti++)
    {
        dttp=*dti;

        if(!dttp->Get_Number_Of_In_Connection())
            dttp->Property()->Set_Offset(0);
        else
            dttp->Property()->Set_Offset(NIL);
        if(!dttp->Get_Number_Of_Out_Connection())
            dttp->Property()->Set_Deadline(dttp->Property()->Get_Global_Deadline());
        else
            dttp->Property()->Set_Deadline(NIL);
    }

    for(dti=dttl.begin();dti!=dttl.end();dti++)
    {
        dttp=*dti;

        path_segment_data_dd psd;

        psd.deadline=dttp->Property()->Get_Deadline();
        psd.pathlen=0;
        psd.distance=0;
        psd.offset=0;
            

        get_path_segment_data(psd,dttp,dsel);
        allocate_slack(psd,dttp,ddp);
        set_deadlines_parents(dttp,psd,ddp);
    }
    
    delete dsel;
    delete rdttl;
    return;
}
        
// Move up the SSSP tree and fill up the get_path_segment_data as we go 
// along. For message deadlines, also include message's latency in the
// distance field.

static void get_path_segment_data (path_segment_data_dd& psd,TaskType * pt,
        list<sssp_entry>* pl)
{
    psd.pathlen++;
    psd.distance+=pt->Property()->Get_Exec_Time();

    if(pt->Property()->Get_Offset()==NIL)
    {
        
        TaskType *pparent=get_parent_task(pt,pl);
        if(pparent) 
        {
            // For messages, add transmission time
            // information.
            
            ConnectionType *pc=pparent->Get_Connection_To(pt);

            psd.distance+=pc->Property()->Get_Trans_Time();
            get_path_segment_data(psd,pparent,pl);
        }
        else psd.offset=0; // Should never happen because of initialization
    }
    else
    {
        psd.offset=pt->Property()->Get_Offset();
        return;
    }
}
    
// Allocate slack based on the info about the path segment and the policy.
// For message deadlines, also allocate deadlines of the message from a 
// task's parents here.

static void allocate_slack(path_segment_data_dd& psd, TaskType *pt, 
        deadline_distribution_policies ddp)
{
    float slack, offset;

    if(pt->Property()->Get_Offset()!=NIL) return;
    
    switch(ddp)
    {
        case NORMAL_DISTRIBUTION:
            slack=((psd.deadline-psd.offset-psd.distance)/psd.distance)*
                    pt->Property()->Get_Exec_Time();
            break;
        case PURE_DISTRIBUTION:
            slack=(psd.deadline-psd.offset-psd.distance)/psd.pathlen;
            break;
    }
    offset=pt->Property()->Get_Deadline()-pt->Property()->Get_Exec_Time()
        -slack;
    pt->Property()->Set_Offset(offset);
    return;
}

// Set the deadlines for the predecessor or parent tasks. If already set, 
// then set it to the minimum of the deadlines. For message deadlines,
// set the deadlines after considering the latency to be allocated to the
// messages from parent node.
    
static void set_deadlines_parents (TaskType *pt, path_segment_data_dd& psd,
        deadline_distribution_policies ddp)
{
    list<ConnectionType *>::iterator sci;

    for(sci=pt->Get_In_Edges_Begin();sci!=pt->Get_In_Edges_End();sci++)
    {
        ConnectionType *scp=*sci;
        float slack_message;

        if(scp->Property()->Get_Trans_Time()>0)
        {
            float psd_distance=psd.distance-pt->Property()->Get_Exec_Time();
			psd.deadline=pt->Property()->Get_Offset();
            
            switch(ddp)
            {
                case NORMAL_DISTRIBUTION:
                    slack_message=((psd.deadline-psd.offset-psd_distance)/psd_distance)*
                       scp->Property()->Get_Trans_Time();
                    break;
                case PURE_DISTRIBUTION:
                    slack_message=(psd.deadline-psd.offset-psd_distance)/psd.pathlen;
                    break;
            }
        }
        else
            slack_message=0;

        float deadline=pt->Property()->Get_Offset()-
            (slack_message+scp->Property()->Get_Trans_Time());
        float pdeadline=scp->Get_From()->Property()->Get_Deadline();
        if((pdeadline==NIL)||(pdeadline>deadline))
            scp->Get_From()->Property()->Set_Deadline(deadline);
    }
    return;
}

