#ifndef FIRST_FIT_TIEBREAKER_H
#define FIRST_FIT_TIEBREAKER_H

//global includes
#include <list>
#include <set>

//local includes
#include "comm_min_graph.h";
#include "processor_network.h"

/*
 * First Fit with a second order tiebreaker mapping
 * ------------------------------------------------
 *
 * This first fit algorithm runs just like the original first fit algorithm with two differneces.
 * First of all, this algorithm allows a check to see if the component that was selected is in 
 * a tie with any other component (or group of components) and then allows a way for this tie to
 * be resolved.  The algorithm goes as follows:
 *
 * 1. Sort all components by utilization (highest utilization first)
 * 2. Sort all processors by the amount of space available on the processor, with processors that
 *    do not have any components assigned to them as lower than all processors that do have
 *    components already assigned to them (see ProcSpaceGreaterThan function object below)
 * 3. For each task t
 *      For each processor p
 *         check to see if utiliztion of t <= space available on p
 *            is t in a tie with any other component or group of components
 *			  if so resolve the tie
 *            assign t or component that resolved the tie to p
 */

//function objects
class ProcSpaceGreaterThan
{
public:
	bool operator()(CommMinNode* node1, CommMinNode* node2)
	{
		double util1 = node1->GetTotalUtilization();
		double util2 = node2->GetTotalUtilization();

		if(util1 && !util2)	//prefer nodes that already have tasks assigned to them
			return true;
		if(util2 && !util1)
			return false;

		//if they both already have tasks assigned to them use the that has the most
		//space
		double space_proc1 = node1->GetAssignedProc()->Property()->Get_Bound() - util1;
		double space_proc2 = node2->GetAssignedProc()->Property()->Get_Bound() - util2;
		return space_proc1 > space_proc2;
	}
};

class UtilizationGreaterThan
{
public:
	bool operator()(CommMinNode* node1, CommMinNode* node2)
	{
		return node1->GetTotalUtilization() > node2->GetTotalUtilization();
	}
};

//Run the first fit with a second order tiebreaker mapping
//This algorithm takes two function pointers as parameters: the first is the function that is called
//when a tie needs to be resolved, the second is used to deteremine if a node is in a tie with another
//node or group of nodes.  The tiebreaker function takes two parameters: the first is a processor node
//that has all of the tasks already assigned to the processor which is where the possible nodes might
//be placed, the second is the list of tied nodes (see below for the format of this list) - this function
//returns the node that breaks the tie, if a group of nodes is determined to break the tie these nodes
//must be merged into a single node when they are returned.  The IsATie function takes 3 parameters: the 
//first if a node that is being checked to see if any other nodes are tied with it, the second is a set 
//of nodes in order of their utilizations that are to be checked to see if a tie has occurred, the third 
//is list of nodes that is populated by the function to store the tied nodes (or group of nodes) that are 
//found - true is returned if any ties are found.

//Format of the tied nodes list: As mentioned before either a single node or a group of nodes can be 
//determined to be in a tie with another node or a group of nodes.  The way that this is allowed in the
//tied nodes list which should be populated in the IsATie function and is passed as a paramter to the
//tiebreaker function is by terminating each group of nodes (may be only a single node) with a NULL 
//character.  This must be taken into consideration when running the tiebreaker function.

//Globals to use: the graph that contains all of the nodes is the first_fit_graph, and the tolerance
//that is used to determine whether or not the utilizations are equal is util_tolerance
double Do_FirstFit_Mapping(CommMinNode* (*tiebreaker)(CommMinNode*, std::list<CommMinNode*>&), bool (*IsATie)(CommMinNode*, std::multiset<CommMinNode*, UtilizationGreaterThan>, std::list<CommMinNode*>&));

//A tie is broken first by removing all groups of nodes that have a node that could still be merged
//to the processor node within the bounds of the given processors bound once the tie has been resolved
//Next the node that saves the greatest communication cost is chosen.  The communication cost is determine
//by the communication cost between nodes in the group (if more than 1 node) + the communication cost
//between the nodes in the group and the processor node - the communication cost between all the nodes in 
//the group and all of the nodes that they are connected to that could not possibly be put onto the
//processor once the tie has been resolved
CommMinNode* CommMinTiebreaker(CommMinNode* node, std::list<CommMinNode*>& tied_nodes);

//A tie for communication minimization is simply when a single node or group of nodes has the equal to
//the utilization of the given node within the constraints of the tolerance
bool CommMinIsATie(CommMinNode* node, std::multiset<CommMinNode*, UtilizationGreaterThan> other_nodes, std::list<CommMinNode*>& tied_nodes);

#endif