#include "first_fit-tiebreaker.h"

//global includes
#include <list>
#include <set>

//local includes
#include "comm_min_graph.h"
#include "processor_network.h"
#include "debug.h"
#include "tools.h"

using namespace std;

/*** Static Helper Functions ***/

// ---Functions for the first fit algorithm

//Assign all of the tasks in the nodes in the global first_fit_graph to the processors 
//that were assigned to these nodes
//Assumptions: all nodes have a processor assigned to them
static void AssignTasksToProcs();

//Add new nodes to the given list of nodes where these new nodes are assigned processors
//that none of the nodes in the given list have been assigned - these newe nodes will have
//no tasks assigned to them
static void AddEmptyNodesWithProcs(list<CommMinNode*>& nodes_with_procs);

//Find the node from the given task that is in the given set of nodes that is not equal to 
//the (optional) not_this_node parameter
static CommMinNode* FindNodeFromTask(TaskType* task, multiset<CommMinNode*, UtilizationGreaterThan>& nodes, CommMinNode* not_this_node = NULL);

//Remove all of the nodes that have any of the tasks that are part of the given node iterator's
//dereferenced value from the given set of nodes and return the iterator position directly after
//the given node iterator
static multiset<CommMinNode*, UtilizationGreaterThan>::iterator EraseAllNodesAt(multiset<CommMinNode*, UtilizationGreaterThan>::iterator node_it, multiset<CommMinNode*, UtilizationGreaterThan>& nodes);


// ---Functions for the communication minimization tiebreaker

//This function checks to see if there are any combinations of nodes where their total utilization
//is equal to the given utilization (taken into consideration the tolerance).  The nodes that are
//checked are in the nodes parameter, the tied_nodes parameter is where the tied nodes are put (list 
//of nodes terminated by NULL)
static void CheckSumsOfUtilizations(multiset<CommMinNode*, UtilizationGreaterThan>& nodes, double utilization, list<CommMinNode*>& tied_nodes);

//This function is a recursive function that is called by CheckSumsOfUtilization.  The other_nodes are
//nodes that are being checked, the utilization is the utilization that still needs to be met (the original
//utilization - the total utilization of the nodes in path), the path parameter are the nodes that have
//been added to the group so far, and the start_it is where in the list of other_nodes should the algorithm
//start checking
static void CheckSumsOfUtilizationsRecurs(multiset<CommMinNode*, UtilizationGreaterThan>& other_nodes, double utilization, list<CommMinNode*>& path, list<CommMinNode*>& tied_nodes, multiset<CommMinNode*, UtilizationGreaterThan>::iterator start_it);

//This function will take out any group of nodes that where once the tie has been resolved (a node was
//chosen and it was merged with the processor node) the group has at least one node where it could still
//be put onto the processor
static void TakeOutGroupsWithSmallNodes(CommMinNode* node, list<CommMinNode*>& tied_nodes);

//This function will remove the given node and all of the following nodes on the node_list until NULL
//is found
static void RemoveGroup(CommMinNode* node, list<CommMinNode*>& node_list);

//Find the node or group of nodes from the given list of tied nodes that saves the most communication
//between the node(s) and the given node, saves the most communication between the group of nodes (if
//a group of nodes is being checked), and causes the least inevitable communication cost between other
//nodes.  Returns the node that does this, or if it is a group of nodes that is chosen, the first node
//of the group is returned
static CommMinNode* FindMinCommCost(CommMinNode* node, list<CommMinNode*>& tied_nodes);

//Find the total communication cost between all connections between the nodes given in the node list
static double FindInnerCommCost(list<CommMinNode*>& node_list);

//Find the total communication cost between all of the nodes in the node_list and the given node
static double FindCommCostToNode(CommMinNode* node, list<CommMinNode*>& node_list);

//Find the total communication cost between all of the nodes in the node_list and all of the nodes that
//they are connected to that: 1) are not the given node and 2) have a utilization that is above max_utilization
static double FindBrokenCost(CommMinNode* node, list<CommMinNode*>& node_list, double max_utilization);

//Find the total utilization of the first group of nodes (may be a single node) terminated by NULL 
static double FindNodeUtilization(list<CommMinNode*>& tied_nodes);

//Find the given node in the list of nodes and merge all nodes (if any) after it until the NULL character
//is found
static void MergeAllNodes(CommMinNode* node, list<CommMinNode*>& nodes);

//function used only for debugging
static void ShowTiedNodes(list<CommMinNode*>& tied_nodes);

//globals
CommMinGraph first_fit_graph;		//graph that makes up all of the nodes
double util_tolerance = .000001;	//tolerance that is used to determine if two nodes
									//have the same utilization
//externals
extern System sys;
extern __int64 num_steps;

double Do_FirstFit_Mapping(CommMinNode* (*tiebreaker)(CommMinNode*, std::list<CommMinNode*>&), bool (*IsATie)(CommMinNode*, std::multiset<CommMinNode*, UtilizationGreaterThan>, std::list<CommMinNode*>&))
{
	list<CommMinNode*> nodes_wo_procs;		//list of nodes that don't have processors
	list<CommMinNode*> nodes_with_procs;	//list of nodes that have processors
	list<CommMinNode*>::iterator node_it;

	multiset<CommMinNode*, ProcSpaceGreaterThan> assigned_nodes;		//set of nodes in order of the original space available on the processor's they were assigned to 
	multiset<CommMinNode*, UtilizationGreaterThan> unassigned_nodes;	//set of nodes in order of their utilization (highest first)

	multiset<CommMinNode*, ProcSpaceGreaterThan>::iterator assigned_node_it;
	multiset<CommMinNode*, UtilizationGreaterThan>::iterator unassigned_node_it;
	multiset<CommMinNode*, UtilizationGreaterThan>::iterator merge_node_it;

	double extra_util = 0;	//the total utilization of all the tasks that
							//could not be assigned to a processor

	num_steps = 0;

	//reset all processors
	for(list<Processor *>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		++num_steps;
		(*proc_it)->Clear_Tasks();							//clear all tasks on the processor
		(*proc_it)->Property()->Reset_Available_Space();	//give the processor its entire amount of space available
	}

	PopulateGraph(&first_fit_graph);			//populate the graph
	MergeNodesWithProcs(&first_fit_graph);		//merge all nodes that have the same processor
												//assigned to them

	GetNodesWithNoProcs(nodes_wo_procs, &first_fit_graph);
	GetNodesWithProcs(nodes_with_procs, &first_fit_graph);
	AddEmptyNodesWithProcs(nodes_with_procs);	//add empty nodes (don't have any tasks) that have a 
												//processor assigned to them where the processor was not
												//represented by any of the nodes in nodes_with_procs

	//order the nodes that have processors
	for(node_it = nodes_with_procs.begin(); node_it != nodes_with_procs.end(); ++node_it)
	{
		++num_steps;
		assigned_nodes.insert(*node_it);
	}

	//order the nodes that do not have processors
	for(node_it = nodes_wo_procs.begin(); node_it != nodes_wo_procs.end(); ++node_it)
	{
		++num_steps;
		unassigned_nodes.insert(*node_it);
	}

	//go through each node with out a processor
	for(unassigned_node_it = unassigned_nodes.begin(); unassigned_node_it != unassigned_nodes.end();)
	{
		CommMinNode* curr_node = *unassigned_node_it;

		//go through each node with a proccessor
		for(assigned_node_it = assigned_nodes.begin(); assigned_node_it != assigned_nodes.end(); ++assigned_node_it)
		{
			CommMinNode* proc_node = *assigned_node_it;

			++num_steps;

			//can the curr_node be put on this processor
			if(proc_node->GetTotalUtilization() + curr_node->GetTotalUtilization() < proc_node->GetAssignedProc()->Property()->Get_Bound())
			{
				list<CommMinNode*> tied_nodes;

				//is this node tied with any other nodes
				if(IsATie(curr_node, unassigned_nodes, tied_nodes))
				{
					tied_nodes.push_front(NULL);
					tied_nodes.push_front(curr_node);				//add curr_node to the list
					curr_node = tiebreaker(proc_node, tied_nodes);	//resolve the tie
				}

				if(curr_node == *unassigned_node_it)	//curr_node was picked to resolve the tie
					unassigned_node_it = EraseAllNodesAt(unassigned_node_it, unassigned_nodes);	//remove the node from the list
				else
				{
					//need to find the node that resolved the tie
					for(merge_node_it = unassigned_node_it; merge_node_it != unassigned_nodes.end(); ++merge_node_it)
						if(*merge_node_it == curr_node) break;

					EraseAllNodesAt(merge_node_it, unassigned_nodes);	//remove the node from the list
				}

				//merge the two nodes
				proc_node->Merge(curr_node);
				first_fit_graph.RemoveNode(curr_node);
				first_fit_graph.RemoveEdgesFrom(curr_node);

				delete curr_node;
				curr_node = NULL;

				break;
			}
		}

		if(curr_node)	//node could not be merged with any of the processor nodes
		{
			extra_util += curr_node->GetTotalUtilization();
			++unassigned_node_it;
		}
	}

	if(extra_util)		//not all of the nodes could be put onto a processor - error
		return extra_util;

	AssignTasksToProcs();

	//print out the results
	print_FirstFit_Mapping();

	return 0;	//no errors
}

static void AssignTasksToProcs()
{
	list<CommMinNode*> nodes;
	list<CommMinNode*>::iterator node_it;
	list<TaskType*>::iterator task_it;

	//go through each node
	nodes = first_fit_graph.GetNodes();
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		++num_steps;

		CommMinNode* node = *node_it;
		Processor* proc = node->GetAssignedProc();
		proc->Property()->Allocate_Space(node->GetTotalUtilization());

		//go through each task that is part of the node
		for(task_it = node->GetTasksBegin(); task_it != node->GetTasksEnd(); ++task_it)
			proc->Add_Task(*task_it);
	}
}

void AddEmptyNodesWithProcs(list<CommMinNode*>& nodes_with_procs)
{
	list<Processor*>::iterator proc_it;
	list<CommMinNode*>::iterator node_it;

	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc = *proc_it;

		++num_steps;

		//find the node that has this processor
		for(node_it = nodes_with_procs.begin(); node_it != nodes_with_procs.end(); ++node_it)
		{
			++num_steps;
			if((*node_it)->GetAssignedProc() == proc) break;
		}

		if(node_it == nodes_with_procs.end())	//no node has the processor - create a new node
		{
			CommMinNode* new_node = new CommMinNode();
			new_node->AssignProc(proc);

			first_fit_graph.AddNode(new_node);
			nodes_with_procs.push_back(new_node);
		}
	}
}

/******* Tie breaker functions *******/
CommMinNode* CommMinTiebreaker(CommMinNode* node, list<CommMinNode*>& tied_nodes)
{
	CommMinNode* best_node;

	TakeOutGroupsWithSmallNodes(node, tied_nodes);			//take out any group of nodes that has a member that could be added once the tie is resolved
	best_node = FindMinCommCost(node, tied_nodes);	//find the node that saves the greatest communication cost

	MergeAllNodes(best_node, tied_nodes);			//merge all of the nodes if a group of nodes was selected to resolve the tie

	return best_node;
}

bool CommMinIsATie(CommMinNode* node, multiset<CommMinNode*, UtilizationGreaterThan> other_nodes, list<CommMinNode*>& tied_nodes)
{
	multiset<CommMinNode*, UtilizationGreaterThan>::iterator node_it;

	double util = node->GetTotalUtilization();
	double curr_util;

	//check to see if any single nodes have an equal utilization
	for(node_it = other_nodes.begin(); node_it != other_nodes.end();)
	{
		++num_steps;

		if(*node_it == node)	//don't worry about the given node
		{
			node_it = other_nodes.erase(node_it);
			continue;
		}

		curr_util = (*node_it)->GetTotalUtilization();

		//check to see if this nodes utilization is in the equal range
		if(util + util_tolerance >= curr_util && curr_util >= util - util_tolerance)
		{
			tied_nodes.push_back(*node_it);
			tied_nodes.push_back(NULL);
			node_it = other_nodes.erase(node_it);
		}
		else
			break;
	}
	
	//check to see if any group of nodes have an equal sum of utilizations
	CheckSumsOfUtilizations(other_nodes, util, tied_nodes);

	if(tied_nodes.size())
		return true;
	return false;
} 

void CheckSumsOfUtilizations(multiset<CommMinNode*, UtilizationGreaterThan>& nodes, double utilization, list<CommMinNode*>& tied_nodes)
{
	list<CommMinNode*> path;
	multiset<CommMinNode*, UtilizationGreaterThan>::iterator it = nodes.begin();

	CheckSumsOfUtilizationsRecurs(nodes, utilization, path, tied_nodes, it);
}

void CheckSumsOfUtilizationsRecurs(multiset<CommMinNode*, UtilizationGreaterThan>& other_nodes, double utilization, list<CommMinNode*>& path, list<CommMinNode*>& tied_nodes, multiset<CommMinNode*, UtilizationGreaterThan>::iterator start_it)
{
	list<CommMinNode*>::iterator path_node_it;
	multiset<CommMinNode*, UtilizationGreaterThan>::iterator node_it;

	//go through each node from the starting position
	for(node_it = start_it; node_it != other_nodes.end(); ++node_it)
	{
		CommMinNode* node = *node_it;
		double curr_util = node->GetTotalUtilization();

		++num_steps;

		//is this nodes utilization below the given utilization + tolerance
		if(util_tolerance + utilization >= curr_util)
		{
			//is this nodes utilization in the acceptable range
			if(curr_util >= utilization - util_tolerance)
			{
				//add the path taken to the tied nodes list - end marked with NULL
				for(path_node_it = path.begin(); path_node_it != path.end(); ++path_node_it)
					tied_nodes.push_back(*path_node_it);
				tied_nodes.push_back(node);
				tied_nodes.push_back(NULL);
			}
			else
			{
				path.push_back(node);	//add the node to the path
				//recursive call - check path starting from the next position and lowering the utilization
				CheckSumsOfUtilizationsRecurs(other_nodes, utilization - curr_util, path, tied_nodes, ++node_it);
				--node_it;	//restore iterator to original position
			}
		}
	}

	if(path.size())			//if there are any nodes on the path remove the last one - this is the position it
		path.pop_back();	//was in before this function was called
}

void TakeOutGroupsWithSmallNodes(CommMinNode* node, list<CommMinNode*>& tied_nodes)
{
	CommMinNode* start_node = NULL;;
	list<CommMinNode*> remove_nodes;
	list<CommMinNode*>::iterator node_it;

	double utilization = FindNodeUtilization(tied_nodes);	//get the utilization of the nodes that are tied

	Processor* proc = node->GetAssignedProc();
	//find the space that will be available on the processor after the tie has been resolved
	double space = proc->Property()->Get_Bound() - (node->GetTotalUtilization() + utilization);

	for(node_it = tied_nodes.begin(); node_it != tied_nodes.end(); ++node_it)
	{
		++num_steps;

		CommMinNode* curr_node = *node_it;
		if(!curr_node)	//NULL marks the end of a group of nodes - reset the start_node and continue
		{
			start_node = NULL;
			continue;
		}

		//if there is more than one node in this group and the current node or the start node can be put
		//onto the processor after the tie is resolved mark the group to be removed
		if(start_node && (curr_node->GetTotalUtilization() <= space || start_node->GetTotalUtilization() <= space))
		{
			remove_nodes.push_back(start_node);

			while(*++node_it != NULL);	//keep going until the end of the group is reached
			start_node = NULL;
			continue;
		}

		if(!start_node)
			start_node = curr_node;
	}

	//remove all groups that were marked to be removed
	for(node_it = remove_nodes.begin(); node_it != remove_nodes.end(); ++node_it)
		RemoveGroup(*node_it, tied_nodes);
}

void RemoveGroup(CommMinNode* remove_node, list<CommMinNode*>& node_list)
{
	//go through each node in the given list
	for(list<CommMinNode*>::iterator node_it = node_list.begin(); node_it != node_list.end(); ++node_it)
	{
		CommMinNode* curr_node = *node_it;

		++num_steps;

		//fond the node
		if(curr_node == remove_node)
		{
			while(*node_it != NULL)		//erase nodes until NULL is found
				node_it = node_list.erase(node_it);
			node_list.erase(node_it);	//erase the NULL
			
			break;
		}
	}
}

CommMinNode* FindMinCommCost(CommMinNode* node, list<CommMinNode*>& tied_nodes)
{
	double cost;
	double best_cost;
	CommMinNode* best_node = NULL;

	list<CommMinNode*> node_list;			//list that contains the group of nodes
	list<CommMinNode*>::iterator node_it;

	double bound = node->GetAssignedProc()->Property()->Get_Bound();

	//find the space that will be available once the tie has been resolved
	double space = bound - (node->GetTotalUtilization() + FindNodeUtilization(tied_nodes));

	//go through each tied node
	for(node_it = tied_nodes.begin(); node_it != tied_nodes.end(); ++node_it)
	{
		CommMinNode* curr_node = *node_it;

		++num_steps;
		
		if(!curr_node)		//end of group of nodes
		{
			cost = FindInnerCommCost(node_list);			//add the cost saved between the nodes in the group
			cost += FindCommCostToNode(node, node_list);	//add the cost saved from the nodes in the group to the processor node 
			cost -= FindBrokenCost(node, node_list, space); //remove the cost from the connections between the nodes
															//in the group and any nodes not in the group that are not the procssor node
															//that have a utilization higher than space
			//save the best node
			if(!best_node || cost > best_cost)
			{
				best_cost = cost;
				best_node = node_list.front();
			}

			//clear the group 
			node_list.clear();
		}
		else
			node_list.push_back(curr_node);	//add node to the group
	}

	return best_node;
}

double FindInnerCommCost(list<CommMinNode*>& node_list)
{
	list<CommMinNode*>::iterator node_it1;
	list<CommMinNode*>::iterator node_it2;

	double comm_cost = 0;

	//go through each node
	for(node_it1 = node_list.begin(); node_it1 != node_list.end(); ++node_it1)
		for(node_it2 = node_it1; node_it2 != node_list.end(); ++node_it2)	//go through each node after the node in
		{																	//the previous loop
			if(node_it1 == node_it2) continue;	//continue past the first position

			++num_steps;

			CommMinNode* node1 = *node_it1;
			CommMinNode* node2 = *node_it2;
			CommMinEdge* edge  = node1->GetEdgeConnecting(node2);	//find the edge between the two

			if(!edge)		//no edge
				continue;

			comm_cost += edge->GetMessageSize();
		}

	return comm_cost;
}

double FindCommCostToNode(CommMinNode* node, list<CommMinNode*>& node_list)
{
	double comm_cost = 0;

	list<CommMinNode*>::iterator node_it;

	//go through each node
	for(node_it = node_list.begin(); node_it != node_list.end(); ++node_it)
	{
		CommMinNode* curr_node = *node_it;
		CommMinEdge* edge  = node->GetEdgeConnecting(curr_node);	//find the edge between the processor node and the
																	//current node
		++num_steps;

		if(!edge)		//no edge
			continue;

		comm_cost += edge->GetMessageSize();
	}

	return comm_cost;
}

double FindBrokenCost(CommMinNode* node, list<CommMinNode*>& node_list, double max_utilization)
{
	double broken_cost = 0;

	list<CommMinNode*>::iterator node_it;
	list<CommMinEdge*>::iterator edge_it;

	//go through each node on the list
	for(node_it = node_list.begin(); node_it != node_list.end(); ++node_it)
	{
		CommMinNode* cur_node = *node_it;

		//go through each node connected to the current node 
		for(edge_it = cur_node->GetEdgeBegin(); edge_it != cur_node->GetEdgeEnd(); ++edge_it)
		{
			CommMinEdge* edge = *edge_it;
			CommMinNode* conn_node = edge->GetOtherNode(cur_node);

			++num_steps;

			if(conn_node == node)	//connected node is equal to the processor node
				continue;

			if(conn_node->GetTotalUtilization() <= max_utilization)	//the utiliziation is below the max
				continue;

			if(OnList<CommMinNode*>(conn_node, node_list))	//this node is part of the group of nodes
				continue;

			broken_cost += edge->GetMessageSize();
		}
	}

	return broken_cost;
}

double FindNodeUtilization(list<CommMinNode*>& tied_nodes)
{
	list<CommMinNode*>::iterator node_it;
	double utilization = 0;

	//go through each node until NULL is found - marks the end of the group
	for(node_it = tied_nodes.begin(); node_it != tied_nodes.end(); ++node_it)
	{
		CommMinNode* curr_node = *node_it;
		if(!curr_node)
			break;

		++num_steps;

		utilization += curr_node->GetTotalUtilization();	//add in the utilization
	}

	return utilization;
}

void MergeAllNodes(CommMinNode* node, list<CommMinNode*>& nodes)
{
	list<CommMinNode*>::iterator node_it;

	//find the position of the given node
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
		if(*node_it == node) break;

	//move to position just after the given node
	++node_it;

	while(1)
	{
		++num_steps;

		CommMinNode* merge_node = *node_it;
		if(!merge_node)	//NULL found - end of group
			break;

		//merge with the first node
		node->Merge(merge_node);
		first_fit_graph.RemoveNode(merge_node);
		first_fit_graph.RemoveEdgesFrom(merge_node);

		//erase from the list and incrament iterator
		node_it = nodes.erase(node_it);
		if(node_it == nodes.end())
			break;
	}
}

multiset<CommMinNode*, UtilizationGreaterThan>::iterator EraseAllNodesAt(multiset<CommMinNode*, UtilizationGreaterThan>::iterator node_it, multiset<CommMinNode*, UtilizationGreaterThan>& nodes)
{
	list<CommMinNode*> erase_nodes;
	multiset<CommMinNode*, UtilizationGreaterThan>::iterator erase_it;
	list<TaskType*>::iterator task_it;
	CommMinNode* node = *node_it;

	//go through each task in the given node
	for(task_it = node->GetTasksBegin(); task_it != node->GetTasksEnd(); ++task_it)
	{
		++num_steps;

		TaskType* task = *task_it;
		CommMinNode* erase_node = FindNodeFromTask(task, nodes, node);	//find the node in the given set of nodes that
																		//has this task that is not the given node
		if(erase_node)	//if the node was found mark it to be erased
			erase_nodes.push_back(erase_node);
	}

	//erase all of the nodes that were marked to be erased
	for(erase_it = nodes.begin(); erase_it != nodes.end();)
	{
		++num_steps;

		CommMinNode* erase_node = *erase_it;
		if(OnList<CommMinNode*>(erase_node, erase_nodes))
		{
			erase_it = nodes.erase(erase_it);
			erase_nodes.remove(erase_node);
		}
		else
			++erase_it;
	}
	
	//erase the given node and return the next position
	return nodes.erase(node_it);
}

CommMinNode* FindNodeFromTask(TaskType* task, multiset<CommMinNode*, UtilizationGreaterThan>& nodes, CommMinNode* not_this_node)
{
	multiset<CommMinNode*, UtilizationGreaterThan>::iterator node_it;
	list<TaskType*>::iterator task_it;

	//go through each node in the list
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		CommMinNode* node = *node_it;

		if(node == not_this_node)
			continue;

		//go through each task that is part of the node
		for(task_it = node->GetTasksBegin(); task_it != node->GetTasksEnd(); ++task_it)
		{
			if(*task_it == task)
				return node;
		}
	}

	return NULL;
}

void ShowTiedNodes(list<CommMinNode*>& tied_nodes)
{
	list<CommMinNode*>::iterator node_it;
	int group = 0;
	CommMinNode* node;
	CString name;

	for(node_it = tied_nodes.begin(); node_it != tied_nodes.end(); ++node_it)
	{ 
		node = *node_it;

		if(!node)
		{
			group++;
			continue;
		}

		name = (*node->GetTasksBegin())->Property()->Get_Name();
	}
}