#include "greedy_b_and_b.h"

//global includes
#include <set>
#include <list>
#include <map>
#include <math.h>

//local includes
#include "task_connection.h"
#include "graph.h"
#include "system.h"
#include "tools.h"
#include "debug.h"

//definitions
#define GBB_INFINITE -1

using namespace std;

//Function Objects
class ProcRankLessThan
{
public:
	bool operator()(Processor* p1, Processor* p2)
	{
		return p1->Property()->Get_GBB_Property()->GetRank() < p2->Property()->Get_GBB_Property()->GetRank();
	}
};

//Helper classes and structs
struct Modifications
{
	list<Processor*> procs_eliminated_from_this;
	list<pair<TaskType*, Processor*> > fc_modified_domains;
	map<Processor*, pair<double, double> >	proc_to_comm_costs;

	double orig_comm_of_assigned_comps;
	double orig_util_of_assigned_comps;
	double orig_mem_of_assigned_comps;
	double orig_mem_cost_of_assigned_comps;
	double orig_comm_cost_of_assigned_comps;

	TaskType* this_task;
};

//External globals
extern Graph grp;
extern System sys;

//Static globals
static list<TaskType*> ordered_tasks;

static double	g_util_of_assigned_comps;
static double	g_mem_of_assigned_comps;
static int		g_mem_cost_of_assigned_comps;
static double	g_ideal_total_mem_cost;
static double	g_comm_of_assigned_comps;
static double	g_comm_cost_of_assigned_comps;	//Mb / second - must be less then g_bus_speed
static double	g_total_comm;

//Globals
double g_util_cost_const = 1;
double g_comm_cost_const = 1;
double g_mem_cost_const  = 1;
double g_bus_speed;			//Mb / second

//Static helper functions
static void		Initialize();
static void		OrderTasks();
static double	GetCommCostTightness(TaskType* task);
static bool		ExpandNextNode(list<TaskType*>::iterator next_task_it, int depth, int& min_fc_level);
static double	EvaluationFunction(TaskType* task, Processor* proc, double added_bus_usage, double total_comm_to_assigned_tasks);
static double	FindExtraCommunication(TaskType* task, Processor* proc, double& total_comm_w_assigned);
static double	FindExtraCommunicationNotOnProc(TaskType* task, Processor* not_proc);
static double	FindTotalCommCost(TaskType* task);
static void		StoreGlobals(Modifications& modifications);
static void		UndoModifications(Modifications& modifications);
static void		RestoreGlobals(Modifications& modifications);
static void		RestoreAllDomains(Modifications& modifications);
static void		RestoreFCDomains(Modifications& modifications);
static void		AssignProc(TaskType* task, Processor* proc, Modifications& modifications);
static void		UnAssignProc(TaskType* task, Processor* proc, Modifications& modifications);
static bool		ForwardCheck(list<TaskType*>::iterator next_task_it, Processor* proc, Modifications& modifications);
static void		SetProcIdealUtilBounds(double total_proc_util);
static void		PermantlyAssignProcs();

double run_greedy_b_and_b()
{
	int min_fc_level = GBB_INFINITE;

	//TODO - DEBUG, BOOKEEPING ERROR WHEN RESTORING DOMAINS - MULTIPLE COPIES OF THE SAME PROC

	Initialize();
	OrderTasks();

	if(ExpandNextNode(ordered_tasks.begin(), 1, min_fc_level))
	{
		PermantlyAssignProcs();
		print_MultipleConstraints_Mapping();
		return 0;
	}
	else
	{
		//Supposed to return the utilization of unallocated tasks - fix later 
		return 1;
	}
}

void Initialize()
{
	list<TaskType*>::iterator task_it;
	list<Processor*>::iterator proc_it;
	list<ConnectionType*>::iterator conn_it;

	g_comm_of_assigned_comps		= 0;
	g_util_of_assigned_comps		= 0;
	g_mem_of_assigned_comps			= 0;
	g_mem_cost_of_assigned_comps	= 0;
	g_comm_cost_of_assigned_comps	= 0;
	g_total_comm					= 0;

	double utilization;
	double total_processor_utilization = 0;
	double total_mem = 0;
	double mem_interval;
	double comm_cost;

	ordered_tasks.clear();

	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc = *proc_it;
		ProcessorGBBProperty* property = proc->Property()->Get_GBB_Property();

		proc->Clear_Tasks();

		if(!property)
		{
			property = new ProcessorGBBProperty;
			proc->Property()->Set_GBB_Property(property);
		}

		property->SetUtilization(0);
		property->SetMemUsed(0);
		property->SetMemCost(0);

		mem_interval = proc->Property()->Get_Mem_Interval();
	}

	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		TaskType* task = *task_it;
		TaskGBBProperty* property = task->Property()->Get_GBB_Property();

		if(!property)
		{
			property = new TaskGBBProperty;
			task->Property()->Set_GBB_Property(property);
		}

		//set the task to be unordered
		property->SetOrdered(false);

		//no processor assigned
		property->SetAssignedProc(NULL);

		property->SetTotalCommCost(FindTotalCommCost(task));

		utilization = task->Property()->Get_Exec_Time() / task->Property()->Get_Period();
		property->SetUtilization(utilization);

		total_processor_utilization += utilization;
		total_mem += task->Property()->Get_Mem_Required();

		//add each processor to the tasks domain
		property->ClearDomain();
		for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
			property->AddToDomain(*proc_it);

		for(conn_it = task->Get_Out_Data_Edges_Begin(); conn_it != task->Get_Out_Data_Edges_End(); ++conn_it)
		{
			ConnectionType* conn = *conn_it;
			comm_cost = conn->Property()->Get_Message_Size() * conn->Property()->Get_Frequency();
			g_total_comm += comm_cost;
		}
	}

	SetProcIdealUtilBounds(total_processor_utilization);
	g_ideal_total_mem_cost = ceil(total_mem/mem_interval);
}

void OrderTasks()
{
	list<TaskType*> unordered_tasks;
	list<TaskType*>::iterator task_it;
	list<TaskType*>::iterator best_task_it;
	list<Processor*>::iterator proc_it;
	double highest_importance;

	Processor* proc;
	ConnectionType* conn;
	TaskType* task;

	double sum_of_util_bounds = 0;
	double sum_of_mem_bounds  = 0;

	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		proc = *proc_it;
		sum_of_util_bounds += proc->Property()->Get_Bound();
		sum_of_mem_bounds  += proc->Property()->Get_Mem_Bound();
	}

	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
		unordered_tasks.push_back(*task_it);

	//loop until all tasks are ordered
	while(unordered_tasks.size())
	{
		best_task_it = unordered_tasks.end();

		//find the unordered task that has the highest importance
		for(task_it = unordered_tasks.begin(); task_it != unordered_tasks.end(); ++task_it)
		{
			task = *task_it;

			double util_importance = g_util_cost_const * task->Property()->Get_GBB_Property()->GetUtilization() / sum_of_util_bounds;
			double mem_importance  = g_mem_cost_const  * task->Property()->Get_Mem_Required() / sum_of_mem_bounds;
			double comm_importance = g_comm_cost_const * GetCommCostTightness(task) / (g_bus_speed*1000000);

			double task_importance = util_importance + mem_importance + comm_importance;

			if(best_task_it == unordered_tasks.end() || task_importance > highest_importance)
			{
				best_task_it = task_it;
				highest_importance = task_importance;
			}
		}

		(*best_task_it)->Property()->Get_GBB_Property()->SetOrdered();
		ordered_tasks.push_back(*best_task_it);
		unordered_tasks.erase(best_task_it);
	}
}

double	GetCommCostTightness(TaskType* task)
{
	list<ConnectionType*>::iterator conn_it;
	bool in_edges = true;
	bool sum_all_edges = false;
	double tightness = 0;

	if(ordered_tasks.empty())
		sum_all_edges = true;

	for(conn_it = task->Get_In_Data_Edges_Begin(); conn_it != task->Get_Out_Data_Edges_End(); ++conn_it)
	{
		if(conn_it == task->Get_In_Data_Edges_End())
		{
			in_edges = false;
			conn_it = task->Get_Out_Data_Edges_Begin();
			if(conn_it == task->Get_Out_Data_Edges_End())
				break;
		}

		ConnectionType* conn = *conn_it;
		TaskType* conn_task;

		if(in_edges)
			conn_task = conn->Get_From();
		else
			conn_task = conn->Get_To();

		if(conn_task->Property()->Get_GBB_Property()->IsOrdered() || sum_all_edges)
		{
			double comm_cost = conn->Property()->Get_Message_Size() * conn->Property()->Get_Frequency();
			tightness += comm_cost;
		}
	}
	return tightness*8 / 1000;
}

bool ExpandNextNode(list<TaskType*>::iterator next_task_it, int depth, int& min_fc_level)
{
	list<Processor*>::iterator proc_it;
	multiset<Processor*, ProcRankLessThan> ordered_procs;
	multiset<Processor*, ProcRankLessThan>::iterator ordered_proc_it;

	Modifications modifications;
	StoreGlobals(modifications);

	TaskType*				task		= *next_task_it;
	TaskGBBProperty*		t_gbb_prop	= task->Property()->Get_GBB_Property();
	ProcessorGBBProperty*	p_gbb_prop;
	Processor*				proc;

	modifications.this_task = task;

	double utilization  = task->Property()->Get_GBB_Property()->GetUtilization();
	double mem_required = task->Property()->Get_Mem_Required();

	double utilization_bound;
	double memory_bound;
	double added_bus_usage;
	double total_comm_to_assigned_tasks=0;

	for(proc_it = t_gbb_prop->GetDomainBegin(); proc_it != t_gbb_prop->GetDomainEnd();)
	{
		bool constraint_violated = false;

		proc		= *proc_it;
		p_gbb_prop	= proc->Property()->Get_GBB_Property();

		utilization_bound	= proc->Property()->Get_Bound();
		memory_bound		= proc->Property()->Get_Mem_Bound();

		added_bus_usage = FindExtraCommunication(task, proc, total_comm_to_assigned_tasks);
		modifications.proc_to_comm_costs[proc] = make_pair(added_bus_usage, total_comm_to_assigned_tasks);

		if(p_gbb_prop->GetMemUsed() + mem_required > memory_bound || p_gbb_prop->GetUtilization() + utilization > utilization_bound)
			constraint_violated = true;
		else if(added_bus_usage + g_comm_of_assigned_comps > g_bus_speed * 1000000)
			constraint_violated = true;

		if(constraint_violated)
		{
			proc_it = t_gbb_prop->EraseFromDomain(proc_it);
			modifications.procs_eliminated_from_this.push_back(proc);

			if(depth < min_fc_level || min_fc_level == GBB_INFINITE)
				min_fc_level = depth;
		}
		else
		{
			p_gbb_prop->SetRank(EvaluationFunction(task, proc, added_bus_usage, total_comm_to_assigned_tasks));
			ordered_procs.insert(proc);
			++proc_it;
		}
	}

	if(t_gbb_prop->GetDomainSize() == 0)
	{
		UndoModifications(modifications);
		return false;
	}

	++next_task_it;

	if(next_task_it == ordered_tasks.end())
	{
		AssignProc(task, *(ordered_procs.begin()), modifications);
		return true;
	}

	for(ordered_proc_it = ordered_procs.begin(); ordered_proc_it != ordered_procs.end(); ++ordered_proc_it)
	{
		bool forward_check_success = true;
		proc = *ordered_proc_it;

		AssignProc(task, proc, modifications);

		if(min_fc_level <= depth && min_fc_level != GBB_INFINITE)
			forward_check_success = ForwardCheck(next_task_it, proc, modifications);
		if(forward_check_success)
		{
			if(ExpandNextNode(next_task_it, depth+1, min_fc_level))
				return true;
		}
		
		UnAssignProc(task, proc, modifications);
	}

	UndoModifications(modifications);
	return false; 
}

double EvaluationFunction(TaskType* task, Processor* proc, double added_comm_cost_of_assigned_tasks, double added_comm_of_assigned_tasks)
{
	double eval_cost = 0;

	TaskGBBProperty*		t_gbb_prop	= task->Property()->Get_GBB_Property();
	ProcessorGBBProperty*	p_gbb_prop	= proc->Property()->Get_GBB_Property();

	//unbalanced utilization cost
	double unbalanced_util_cost = (t_gbb_prop->GetUtilization() + p_gbb_prop->GetUtilization() - p_gbb_prop->GetIdealUtil()) / p_gbb_prop->GetIdealUtil();

	//memory cost
	double mem_required			= task->Property()->Get_Mem_Required();
	double new_proc_mem			= p_gbb_prop->GetMemUsed() + mem_required;
	double added_mem_cost		= ceil(new_proc_mem / proc->Property()->Get_Mem_Interval()) - p_gbb_prop->GetMemCost();
	double mem_cost				= (added_mem_cost + g_mem_cost_of_assigned_comps - g_ideal_total_mem_cost) / g_ideal_total_mem_cost;

	//communication cost
	double comm_cost;
	if(g_total_comm)
		comm_cost = (g_comm_cost_of_assigned_comps + added_comm_cost_of_assigned_tasks - g_total_comm) / g_total_comm;
	else
		comm_cost = 0;

	eval_cost = g_util_cost_const*unbalanced_util_cost + g_mem_cost_const*mem_cost + g_comm_cost_const*comm_cost;
	
	return eval_cost;
}

double FindExtraCommunication(TaskType* task, Processor* proc, double& total_comm_w_assigned)
{
	list<ConnectionType*>::iterator conn_it;

	bool					in_edges = true;
	TaskType*				conn_task;
	ConnectionType*			conn;
	Processor*				conn_proc;
	double added_bus_usage = 0;
	double edge_cost;

	total_comm_w_assigned = 0;

	for(conn_it = task->Get_In_Data_Edges_Begin(); conn_it != task->Get_Out_Data_Edges_End(); ++conn_it)
	{
		if(conn_it == task->Get_In_Data_Edges_End())
		{
			conn_it = task->Get_Out_Data_Edges_Begin();
			if(conn_it == task->Get_Out_Data_Edges_End())
				break;

			in_edges = false;
		}

		conn = *conn_it;

		if(in_edges)
			conn_task = conn->Get_From();
		else
			conn_task = conn->Get_To();

		conn_proc = conn_task->Property()->Get_GBB_Property()->GetAssignedProc();
		if(conn_proc)
		{	
			edge_cost = (conn->Property()->Get_Message_Size() * 8) / (conn->Property()->Get_Frequency() * 1000);
			
			total_comm_w_assigned += edge_cost;
			if(conn_proc != proc)
				added_bus_usage += edge_cost;
		}
	}

	return added_bus_usage;
}

double FindExtraCommunicationNotOnProc(TaskType* task, Processor* not_proc)
{
	list<ConnectionType*>::iterator conn_it;

	bool					in_edges = true;
	TaskType*				conn_task;
	ConnectionType*			conn;
	Processor*				conn_proc;
	double added_bus_usage = 0;
	double edge_cost;

	for(conn_it = task->Get_In_Data_Edges_Begin(); conn_it != task->Get_Out_Data_Edges_End(); ++conn_it)
	{
		if(conn_it == task->Get_In_Data_Edges_End())
		{
			conn_it = task->Get_Out_Data_Edges_Begin();
			if(conn_it == task->Get_Out_Data_Edges_End())
				break;

			in_edges = false;
		}

		conn = *conn_it;

		if(in_edges)
			conn_task = conn->Get_From();
		else
			conn_task = conn->Get_To();

		conn_proc = conn_task->Property()->Get_GBB_Property()->GetAssignedProc();
		if(conn_proc)
		{	
			edge_cost = (conn->Property()->Get_Message_Size() * 8) / (conn->Property()->Get_Frequency() * 1000);
			
			if(conn_proc == not_proc)
				added_bus_usage += edge_cost;
		}
	}

	return added_bus_usage;
}

double FindTotalCommCost(TaskType* task)
{
	double comm_cost = 0;
	list<ConnectionType*>::iterator conn_it;

	for(conn_it = task->Get_In_Data_Edges_Begin(); conn_it != task->Get_Out_Data_Edges_End(); ++conn_it)
	{
		if(conn_it == task->Get_In_Data_Edges_End())
		{
			conn_it = task->Get_Out_Data_Edges_Begin();
			if(conn_it == task->Get_Out_Data_Edges_End())
				break;
		}

		comm_cost += ((*conn_it)->Property()->Get_Message_Size() * 8) / ((*conn_it)->Property()->Get_Frequency() * 1000);
	}

	return comm_cost;
}

void StoreGlobals(Modifications& modifications)
{
	modifications.orig_comm_cost_of_assigned_comps	= g_comm_cost_of_assigned_comps;
	modifications.orig_comm_of_assigned_comps		= g_comm_of_assigned_comps;
	modifications.orig_mem_cost_of_assigned_comps	= g_mem_cost_of_assigned_comps;
	modifications.orig_mem_of_assigned_comps		= g_mem_of_assigned_comps;
	modifications.orig_util_of_assigned_comps		= g_util_of_assigned_comps;
}

void UndoModifications(Modifications& modifications)
{
	RestoreGlobals(modifications);
	RestoreAllDomains(modifications);
}

void RestoreGlobals(Modifications& modifications)
{
	g_comm_cost_of_assigned_comps	= modifications.orig_comm_cost_of_assigned_comps;
	g_comm_of_assigned_comps		= modifications.orig_comm_of_assigned_comps;
	g_mem_cost_of_assigned_comps	= modifications.orig_mem_cost_of_assigned_comps;
	g_mem_of_assigned_comps			= modifications.orig_mem_of_assigned_comps;
	g_util_of_assigned_comps		= modifications.orig_util_of_assigned_comps;
}

void RestoreAllDomains(Modifications& modifications)
{
	list<Processor*>::iterator proc_it;
	TaskGBBProperty* t_gbb_prop = modifications.this_task->Property()->Get_GBB_Property();

	RestoreFCDomains(modifications);

	for(proc_it = modifications.procs_eliminated_from_this.begin(); proc_it != modifications.procs_eliminated_from_this.end(); ++proc_it)
		t_gbb_prop->AddToDomain(*proc_it);
}

void RestoreFCDomains(Modifications& modifications)
{
	list<pair<TaskType*, Processor*> >::iterator it;

	for(it = modifications.fc_modified_domains.begin(); it != modifications.fc_modified_domains.end(); ++it)
		it->first->Property()->Get_GBB_Property()->AddToDomain(it->second);

	modifications.fc_modified_domains.clear();
}

void AssignProc(TaskType* task, Processor* proc, Modifications& modifications)
{
	TaskGBBProperty*		t_gbb_prop	= task->Property()->Get_GBB_Property();
	ProcessorGBBProperty*	p_gbb_prop	= proc->Property()->Get_GBB_Property();

	//Modify Processor
	int orig_mem_cost = p_gbb_prop->GetMemCost();
	p_gbb_prop->AddMemUsed(task->Property()->Get_Mem_Required());
	int new_mem_cost = ceil(p_gbb_prop->GetMemUsed() / proc->Property()->Get_Mem_Interval());
	p_gbb_prop->SetMemCost(new_mem_cost);
	p_gbb_prop->AddUtilization(t_gbb_prop->GetUtilization());

	//Modify Globals
	g_mem_cost_of_assigned_comps	+= (new_mem_cost - orig_mem_cost);
	g_mem_of_assigned_comps			+= task->Property()->Get_Mem_Required();
	g_comm_cost_of_assigned_comps	+= modifications.proc_to_comm_costs[proc].first;
	g_comm_of_assigned_comps		+= modifications.proc_to_comm_costs[proc].second;
	g_util_of_assigned_comps		+= t_gbb_prop->GetUtilization();

	//Modify Task
	t_gbb_prop->SetAssignedProc(proc);
}

void UnAssignProc(TaskType* task, Processor* proc, Modifications& modifications)
{
	TaskGBBProperty*		t_gbb_prop	= task->Property()->Get_GBB_Property();
	ProcessorGBBProperty*	p_gbb_prop	= proc->Property()->Get_GBB_Property();

	RestoreFCDomains(modifications);
	RestoreGlobals(modifications);

	//Restore processor to state before assignment
	p_gbb_prop->RemoveMemUsed(task->Property()->Get_Mem_Required());
	p_gbb_prop->SetMemCost(p_gbb_prop->GetMemUsed() / proc->Property()->Get_Mem_Interval());
	p_gbb_prop->RemoveUtilization(t_gbb_prop->GetUtilization());

	//Assign the processor to be NULL
	t_gbb_prop->SetAssignedProc(NULL);
}

bool ForwardCheck(list<TaskType*>::iterator next_task_it, Processor* proc, Modifications& modifications)
{
	double added_bus_usage;
	double utilization_bound			= proc->Property()->Get_Bound();
	double memory_bound					= proc->Property()->Get_Mem_Bound();
	double total_comm_to_assigned_tasks	= 0;
	ProcessorGBBProperty* p_gbb_prop	= proc->Property()->Get_GBB_Property();

	list<Processor*>::iterator proc_it;

	for(; next_task_it != ordered_tasks.end(); ++next_task_it)
	{
		bool constraint_violated	= false;
		bool in_domain				= false;

		TaskType*			task		= *next_task_it;
		TaskGBBProperty*	t_gbb_prop	= task->Property()->Get_GBB_Property();

		if(t_gbb_prop->InDomain(proc))
		{
			in_domain = true;

			added_bus_usage = FindExtraCommunication(task, proc, total_comm_to_assigned_tasks);
		
			if(p_gbb_prop->GetMemUsed() + task->Property()->Get_Mem_Required() > memory_bound || p_gbb_prop->GetUtilization() + t_gbb_prop->GetUtilization() > utilization_bound)
				constraint_violated = true;
			else if(added_bus_usage + g_comm_of_assigned_comps > g_bus_speed * 1000000)
				constraint_violated = true;

			if(constraint_violated)
			{
				t_gbb_prop->RemoveFromDomain(proc);
				modifications.fc_modified_domains.push_back(make_pair(task, proc));
			}
		}

		if((in_domain && t_gbb_prop->GetDomainSize() > 1) || ((!in_domain || constraint_violated) && t_gbb_prop->GetDomainSize() == 1))
		{
			added_bus_usage = FindExtraCommunicationNotOnProc(task, proc);
		
			if(added_bus_usage + g_comm_of_assigned_comps > g_bus_speed * 1000000)
			{
				for(proc_it = t_gbb_prop->GetDomainBegin(); proc_it != t_gbb_prop->GetDomainEnd();)
				{
					if(*proc_it == proc)
					{
						++proc_it;
						continue;
					}

					modifications.fc_modified_domains.push_back(make_pair(task, *proc_it));
					proc_it = t_gbb_prop->EraseFromDomain(proc_it);
				}
			}
		}

		if(t_gbb_prop->GetDomainSize() == 0)
			return false;
	}

	return true;
}

void SetProcIdealUtilBounds(double total_proc_util)
{
	list<Processor*> avail_procs;
	list<Processor*>::iterator proc_it;
	Processor* proc;
	ProcessorGBBProperty* p_gbb_prop;
	double avg_util;
	double util_bound;
	double new_util;

	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
		avail_procs.push_back(*proc_it);

	while(total_proc_util > .0000000001 && avail_procs.size())
	{
		avg_util = total_proc_util/avail_procs.size();

		for(proc_it = avail_procs.begin(); proc_it != avail_procs.end();)
		{
			proc = *proc_it;

			p_gbb_prop = proc->Property()->Get_GBB_Property();
			util_bound = proc->Property()->Get_Bound();

			new_util   = p_gbb_prop->GetIdealUtil() + avg_util;

			if(new_util <= util_bound)
			{
				p_gbb_prop->SetIdealUtil(new_util);
				total_proc_util -= avg_util;
				++proc_it;
			}
			else
			{
				p_gbb_prop->SetIdealUtil(util_bound);
				total_proc_util -= (new_util - util_bound);
				proc_it = avail_procs.erase(proc_it);
			}

		}
	}
}

void PermantlyAssignProcs()
{
	list<TaskType*>::iterator task_it;
	TaskType* task;
	Processor* proc;

	for(task_it = ordered_tasks.begin(); task_it != ordered_tasks.end(); ++task_it)
	{
		task = *task_it;
		proc = task->Property()->Get_GBB_Property()->GetAssignedProc();
		proc->Property()->Allocate_Space(task->Property()->Get_GBB_Property()->GetUtilization());
		proc->Add_Task(task);
	}
}