#include "optimal_mapping.h"

//global includes
#include <list>
#include <map>

//local includes
#include "system.h"
#include "graph.h"
#include "tools.h"
#include "debug.h"

using namespace std;

//externals
extern __int64 num_steps;
extern Graph grp;	
extern System sys;

//globals
double g_best_comm_cost;
double g_best_balanced_load_cost;
double g_best_first_fit_cost;

map<Processor*, double> proc_util_for_comm;
map<Processor*, double> proc_util_for_load_balance;
map<Processor*, double> proc_util_for_first_fit;

double load_balance_comm_cost;
double first_fit_comm_cost;

//static helper functions
static void InitOptimalMapping();
static void OptimalMappingExpand(list<TaskType*>::iterator node_it);
static void FindAndStoreCosts();
static double FindCommCost(bool find_best=true);
static double FindFirstFitCost();
static double FindLoadBalanceCost();
static void StoreProcUtil(map<Processor*, double>& proc_util_map);
 
double do_optimal_mapping()
{
	InitOptimalMapping();

	OptimalMappingExpand(grp.Node_Begin());	//expand the first node

	if(g_best_comm_cost == -1)			//all costs == -1 if there were no
		return 1;						//legal assignemts for every task

	print_Optimal_Mapping();

	return 0;
}

void InitOptimalMapping()
{
	list<TaskType*>::iterator  task_it;
	list<Processor*>::iterator proc_it;

	num_steps					= 0;		//reset the number of steps taken

	g_best_comm_cost			= -1;	
	g_best_balanced_load_cost	= -1;
	g_best_first_fit_cost		= -1;

	proc_util_for_comm.clear();
	proc_util_for_load_balance.clear();
	proc_util_for_first_fit.clear();

	load_balance_comm_cost		= -1;
	first_fit_comm_cost			= -1;

	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		TaskType* task = *task_it;
		TaskGBBProperty* property = task->Property()->Get_GBB_Property();

		if(!property)
		{
			property = new TaskGBBProperty;
			task->Property()->Set_GBB_Property(property);
		}

		//no processor assigned
		property->SetAssignedProc(NULL);

		property->SetUtilization(task->Property()->Get_Exec_Time() / task->Property()->Get_Period());
	}
	
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc = *proc_it;
		ProcessorGBBProperty* property = proc->Property()->Get_GBB_Property();

		proc->Clear_Tasks();

		if(!property)
		{
			property = new ProcessorGBBProperty;
			proc->Property()->Set_GBB_Property(property);
		}

		property->SetUtilization(0);
	}

}

void OptimalMappingExpand(list<TaskType*>::iterator node_it)
{
	list<Processor*>::iterator proc_it;

	TaskType* task = *node_it;
	TaskGBBProperty* t_gbb_prop = task->Property()->Get_GBB_Property();
	ProcessorGBBProperty* p_gbb_prop;

	double util = t_gbb_prop->GetUtilization();

	//try each processor for the current task
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		++num_steps;		//increment number of steps taken

		Processor* proc = *proc_it;
		p_gbb_prop = proc->Property()->Get_GBB_Property();

		//can this task be assigned to this proc without violating its bound
		if(p_gbb_prop->GetUtilization() + util > proc->Property()->Get_Bound())
			continue;

		p_gbb_prop->AddUtilization(util);	//updated utilziation
		t_gbb_prop->SetAssignedProc(proc);	//set the processor

		if(++node_it == grp.Node_End())
			FindAndStoreCosts();
		else
			OptimalMappingExpand(node_it);	//move on to the next task

		--node_it;

		//reset to previous value before assignment
		p_gbb_prop->RemoveUtilization(util);
	}
}

void FindAndStoreCosts()
{
	double comm_cost;
	double first_fit_cost;
	double load_balance_cost;

	comm_cost = FindCommCost();
	if(comm_cost != -1 && (comm_cost < g_best_comm_cost || g_best_comm_cost == -1))
	{
		StoreProcUtil(proc_util_for_comm);
		g_best_comm_cost = comm_cost;
	}

	first_fit_cost = FindFirstFitCost();
	if(first_fit_cost != -1 && (first_fit_cost > g_best_first_fit_cost || g_best_first_fit_cost == -1))
	{
		first_fit_comm_cost = FindCommCost(false);
		StoreProcUtil(proc_util_for_first_fit);
		g_best_first_fit_cost = first_fit_cost;
	}

	load_balance_cost = FindLoadBalanceCost();
	if(load_balance_cost != -1 && (load_balance_cost < g_best_balanced_load_cost || g_best_balanced_load_cost == -1))
	{
		load_balance_comm_cost = FindCommCost(false);
		StoreProcUtil(proc_util_for_load_balance);
		g_best_balanced_load_cost = load_balance_cost;
	}
}

double FindCommCost(bool find_best)
{
	double comm_cost = 0;

	list<TaskType*>::iterator task_it;
	list<ConnectionType*>::iterator conn_it;
	TaskType* task;
	Processor* proc;
	ConnectionType* conn;

	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		task = *task_it;
		proc = task->Property()->Get_GBB_Property()->GetAssignedProc();

		for(conn_it = task->Get_Out_Data_Edges_Begin(); conn_it != task->Get_Out_Data_Edges_End(); ++conn_it)
		{
			conn = *conn_it;
			if(conn->Get_To()->Property()->Get_GBB_Property()->GetAssignedProc() != proc)
			{
				comm_cost += conn->Property()->Get_Message_Size() * conn->Property()->Get_Frequency();
				if(find_best && comm_cost > g_best_comm_cost && g_best_comm_cost != -1)
					return -1;
			}
		}
	}

	return comm_cost;
}

double FindFirstFitCost()
{
	double first_fit_cost = 0;

	list<Processor*>::iterator proc_it;
	ProcessorGBBProperty* p_gbb_prop;

	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		p_gbb_prop = (*proc_it)->Property()->Get_GBB_Property();
		first_fit_cost += p_gbb_prop->GetUtilization() * p_gbb_prop->GetUtilization();
	}

	return first_fit_cost;
}

double FindLoadBalanceCost()
{
	double load_balance_cost = 0;

	list<Processor*>::iterator proc_it1;
	list<Processor*>::iterator proc_it2;

	ProcessorGBBProperty* p_gbb_prop1;
	ProcessorGBBProperty* p_gbb_prop2;

	for(proc_it1 = sys.Proc_Begin(); proc_it1 != sys.Proc_End(); ++proc_it1)
	{
		for(proc_it2 = proc_it1; proc_it2 != sys.Proc_End(); ++proc_it2)
		{
			if(proc_it2 == proc_it1)
			{
				if(++proc_it2 == sys.Proc_End())
					break;
			}

			p_gbb_prop1 = (*proc_it1)->Property()->Get_GBB_Property();
			p_gbb_prop2 = (*proc_it2)->Property()->Get_GBB_Property();

			load_balance_cost += abs_double(p_gbb_prop1->GetUtilization() - p_gbb_prop2->GetUtilization());
		
			if(load_balance_cost > g_best_balanced_load_cost && g_best_balanced_load_cost != -1)
				return -1;
		}
	}

	return load_balance_cost;
}

void StoreProcUtil(map<Processor*, double>& proc_util_map)
{
	list<Processor*>::iterator proc_it;
	Processor* proc;
	ProcessorGBBProperty* p_gbb_prop;

	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		proc = *proc_it;
		p_gbb_prop = proc->Property()->Get_GBB_Property();
		proc_util_map[proc] = p_gbb_prop->GetUtilization();
	}
}