#include "..\stdafx.h"
#include "windows.h"
#include <stdio.h>
#include <math.h>
#include <algorithm>
#include <map>
#include <utility>
#include <assert.h>
#include "graphalgo.h"
#include "system.h"
#include "treenode.h"
#include "task_allocation.h"
#include "schedule.h"
#include "debug.h"
#include "priority_assign.h"
#include "main_extern.h"

typedef struct
{
    TaskType* ptask;
    int priority;
} task_priority_data;

static map<int, TaskType*> taskidmap;
static set<int> rand_int_list;

static bool get_energy_value (float&,Processor*,SA_parameters&);
static void save_task_priority (list<task_priority_data>& ltpd, Processor *pproc);
static void restore_task_priority (list<task_priority_data>&);
static void initialize (Processor*);
static int get_direction_of_move();
static TaskType* get_task_to_move (int&);
static void move (task_priority_data&);
static void undo_move (task_priority_data&);
static bool thermal_equillibrium (SA_statistics&,SA_parameters&);
static bool termination_condition(bool&,SA_statistics&,SA_parameters&);
static bool priority_lowerable (TaskType*);
static bool priority_raisable (TaskType*);
static void order_tasks_list (list<TaskType*>&,list<TaskType*>&);
static float get_new_temperature(SA_parameters&,float);
static int get_priority_to_move (TaskType*,int);
static float get_initial_temperature (Processor*,SA_parameters&);
static void unique_rand_init();
static int unique_rand(int);
static void initialize_sa_parameters(Processor*,SA_parameters&);
static bool check_utilization (Processor*);
static void algo_clear ();

void SA_priority_assign (Processor *pproc, SA_parameters& sa)
{
    SA_statistics st;
    list<task_priority_data> best_sol;
    float en,best_val;
    bool bn;

	algo_clear();
    st.num_jumps=st.lower_energy_jumps=st.higher_energy_jumps=0;

    if(!check_utilization(pproc)) return;
    initialize(pproc);
    bn=get_energy_value(en,pproc,sa);
    initialize_sa_parameters(pproc, sa);
    float temperature=sa.initial_temperature;

    if(bn) 
    {
        best_val=en;
        save_task_priority(best_sol,pproc);
    }
    else best_val=99999999999;

    while(1)
    {
        task_priority_data md;
        float em;
        
        st.num_jumps++;
        
        move(md);
        bool bm=get_energy_value(em,pproc,sa);

        if(bm&&(em<best_val))
        {
            best_val=em;
            save_task_priority(best_sol,pproc);
        }
        
        if((bm&&((em<en)||!bn))||((!bm)&&(!bn)&&(em<en)))
        {
            st.lower_energy_jumps++;
            en=em;
            bn=bm;
        }
        else
        {
            if((!bm)&&(bn)&&(em>=en))
            {
                undo_move(md);
                Restore_MP_SP(pproc);
            }
            else
            {
                float diff_energy=en-em;
                diff_energy=(diff_energy>0)?(-diff_energy):diff_energy;
                float e=exp(diff_energy/temperature);
                float r=rand()/(RAND_MAX+1.0);
                if(e>r) 
                {
                    st.higher_energy_jumps++;
                    en=em;
                    bn=bm;
                }
                else
                {
                    undo_move(md);
                    Restore_MP_SP(pproc);
                }
            }
        }
        if(termination_condition(bn, st, sa)) break;
        if(thermal_equillibrium(st, sa))
            temperature=get_new_temperature(sa, temperature);

        if(st.num_jumps%sa.equillibrium_interval==0)
        {
            st.lower_energy_jumps=st.higher_energy_jumps=0;
        }

        printf("\n%d: En %.3f Sol %d T %.3f LJ %d HJ %d",st.num_jumps,en,bn?1:0,temperature,st.lower_energy_jumps,st.higher_energy_jumps);
    }
    bn=get_energy_value(en,pproc,sa);
    if(best_val<en) 
    {
        restore_task_priority(best_sol);
        get_energy_value(en,pproc,sa);
    }
    return;
}

static void initialize_sa_parameters (Processor *pproc, SA_parameters& sa)
{
    list<TaskType*>::iterator ti;
    float max_deadline=-999999;
    int num_levels=taskidmap.size();
    int roots=0;
    
    for(ti=pproc->Tasks_Begin();ti!=pproc->Tasks_End();ti++)
    {
        TaskType* pt=*ti;

        if(pt==pt->Property()->Get_Global_Root())
        {
            if(pt->Property()->Get_Global_Deadline()>max_deadline)
                max_deadline=pt->Property()->Get_Global_Deadline();
            roots++;
        }
    }
    sa.initial_temperature=(20.0*(sa.min_deadline_constant*max_deadline+
                sa.distinct_level_constant*(num_levels-roots)));
    sa.equillibrium_interval=max(num_levels*roots,20);
    sa.equillibrium_jumps=max(8,(int)0.4*sa.equillibrium_interval);
}

static bool thermal_equillibrium (SA_statistics& st, SA_parameters& sa)
{
    if(st.lower_energy_jumps>=sa.equillibrium_jumps) 
    {
        st.lower_energy_jumps=0;
        st.higher_energy_jumps=0;
        return true;
    }
    if(st.num_jumps%sa.equillibrium_interval==0)
    {
        st.lower_energy_jumps=0;
        st.higher_energy_jumps=0;
        return true;
    }
    return false;
}

static bool termination_condition(bool& solution, SA_statistics& st, SA_parameters& sa)
{
    if(st.num_jumps%sa.equillibrium_interval==0)
    {
        if(solution&&(st.lower_energy_jumps==0)&&(st.higher_energy_jumps==0)) return true;
    }
    if(st.num_jumps>sa.stopping_jumps) return true;
    else return false;
}

static void save_task_priority (list<task_priority_data>& ltpd, Processor *pproc)
{
    list<TaskType*>::iterator ti;

    ltpd.clear();
    for(ti=pproc->Tasks_Begin();ti!=pproc->Tasks_End();ti++)
    {
        TaskType* pt=*ti;

        task_priority_data tpd;

        tpd.ptask=pt;
        tpd.priority=pt->Property()->Get_Priority();
        ltpd.push_front(tpd);
    }
}

static void restore_task_priority (list<task_priority_data>& ltpd)
{
    list<task_priority_data>::iterator ti;

    for(ti=ltpd.begin();ti!=ltpd.end();ti++)
    {
        TaskType* pt=(*ti).ptask;

        pt->Property()->Set_Priority((*ti).priority);
    }
}

bool get_energy_value (float& ret, Processor* pproc, SA_parameters& sa)
{
    set<int> levels;
    list<TaskType*>::iterator ti;
    float mind=-99999999;
    int sum_thread_comm=0;
    
    Schedule_Task(pproc);
    for(ti=pproc->Tasks_Begin();ti!=pproc->Tasks_End();ti++)
    {
        TaskType* pt=*ti;

        if(pt->Property()->Get_Global_Root()==pt)
        {
            float delay=pt->Property()->Get_Global_Response_Time()-
                pt->Property()->Get_Global_Deadline();
            if(delay>mind) mind=delay;

            // Get inter thread communication

            list<TaskType*>* topol=pt->Property()->Topo();
            list<TaskType*>::iterator tpi;
            TaskType* pt1=pt;
            int stc=0;
            
			tpi=topol->begin();tpi++;
            for(;tpi!=topol->end();tpi++)
            {
                TaskType* pt2=*tpi;

                if(pt1->Property()->Get_Priority()!=pt2->Property()->Get_Priority())
                    stc++;
                pt1=pt2;
            }
            sum_thread_comm+=stc;
        }
        levels.insert(pt->Property()->Get_Priority());
    }

    ret=sa.min_deadline_constant*mind+sa.distinct_level_constant*levels.size()
        +sa.thread_comm_constant*sum_thread_comm;
    return mind<=0;
}

static void initialize (Processor* pproc)
{
    list<TaskType*>::iterator ti;
    list<TaskType*> roots, ordered_roots;
    int i=0;

    for(ti=pproc->Tasks_Begin();ti!=pproc->Tasks_End();ti++)
    {
        TaskType *pt=*ti;

        if(pt->Property()->Get_Global_Root()==pt) roots.push_front(pt);

        pair<int, TaskType*> tmp_pair;

        tmp_pair.second=pt;
        tmp_pair.first=i++;

        taskidmap.insert(tmp_pair);
    }
    
    order_tasks_list(ordered_roots, roots);

    int input_priority=0;
    for(ti=grp.Node_Begin();ti!=grp.Node_End();ti++)
    {
        TaskType* pt=*ti;
        if(pt->Property()->Get_User_Priority()>input_priority)
            input_priority=pt->Property()->Get_User_Priority();
    }
    
    int priority=max(input_priority+1,taskidmap.size());

    for(ti=ordered_roots.begin();ti!=ordered_roots.end();ti++)
    {
        TaskType* pt=*ti;

        list<TaskType*>* topol=pt->Property()->Topo();
        list<TaskType*>::iterator ti2;

        for(ti2=topol->begin();ti2!=topol->end();ti2++)
        {
            TaskType* pt2=*ti2;

            if(pt2->Property()->Get_User_Priority()==NIL)
                pt2->Property()->Set_Priority(priority);
            else
                pt2->Property()->Set_Priority(pt2->Property()->Get_User_Priority());
        }
        priority=priority-topol->size();
    }
}

static int get_direction_of_move ()
{
    return (int)(2.0*rand()/(RAND_MAX+1.0));
}

static TaskType* get_task_to_move (int& dir)
{
    list<TaskType*>* topol;
    list<TaskType*>::reverse_iterator ri;
    list<TaskType*>::iterator ti;
    set<TaskType*> tried_tasks;
    int size=taskidmap.size();
    TaskType* ptask=NULL;
    
    unique_rand_init();

    while(1)
    {
        
        int rand_root=unique_rand(size);
        map<int, TaskType*>::iterator mi=taskidmap.find(rand_root);
        ptask=mi->second;

        set<TaskType*>::iterator si=tried_tasks.find(ptask);
        if(si!=tried_tasks.end())
        {
            if(tried_tasks.size()==taskidmap.size())
            {
                tried_tasks.clear();
                dir=(dir+1)%2;
                unique_rand_init();
                if(taskidmap.size()==1) 
                    return (ptask->Property()->Get_User_Priority()==NIL)?ptask:NULL;
                continue;
            }
            else continue;
        }
        else
            tried_tasks.insert(ptask);

        switch(dir)
        {
            case 0: // Move priority of tasks down;
                topol=ptask->Property()->Topo();

                ri=find(topol->rbegin(),topol->rend(),ptask);

                for(;ri!=topol->rend();ri++)
                {
                    TaskType* pt=*ri;
                    if(priority_lowerable(pt)) 
                        return (pt->Property()->Get_User_Priority()==NIL)?pt:NULL;
                }
                break;
            case 1: // Move priority of tasks up;
                topol=ptask->Property()->Topo();

                ti=find(topol->begin(),topol->end(),ptask);

                for(;ti!=topol->end();ti++)
                {
                    TaskType* pt=*ti;
                    if(priority_raisable(pt)) 
                        return (pt->Property()->Get_User_Priority()==NIL)?pt:NULL;
                }
                break;
        }
    }
}
            
static bool priority_lowerable (TaskType* pt)
{
    int pr=pt->Property()->Get_Priority();
    
    if(pr==0) return false;
    list<ConnectionType*>::iterator ci;

    if(pt->Get_Number_Of_In_Connection()==0) return true;
    for(ci=pt->Get_In_Edges_Begin();ci!=pt->Get_In_Edges_End();ci++)
    {
        ConnectionType* pc=*ci;
        TaskType* pfrom=pc->Get_From();

        if(pfrom->Property()->Get_Priority()<pr) return true;
    }
    return false;
}

static bool priority_raisable (TaskType* pt)
{
    int pr=pt->Property()->Get_Priority();

    if(pr==(taskidmap.size()-1)) return false;
    list<ConnectionType*>::iterator ci;

    if(pt->Get_Number_Of_Out_Connection()==0) return true;
    for(ci=pt->Get_Out_Edges_Begin();ci!=pt->Get_Out_Edges_End();ci++)
    {
        ConnectionType* pc=*ci;
        TaskType* pto=pc->Get_To();

        if(pto->Property()->Get_Priority()>pr) return true;
    }
    return false;
}

static void move (task_priority_data& md)
{
    int dir=get_direction_of_move();
    TaskType* pt=get_task_to_move(dir);

    if(pt==NULL) 
    {
        md.ptask=NULL;
        return;
    }
    
    int pr=get_priority_to_move(pt, dir);

    md.ptask=pt;
    md.priority=pt->Property()->Get_Priority();

    pt->Property()->Set_Priority(pr);
    return;
}

static void undo_move (task_priority_data& md)
{
    TaskType* pt=md.ptask;

    if(pt==NULL) return;

    int pr=md.priority;

    pt->Property()->Set_Priority(pr);
    return;
}
    
static int get_priority_to_move (TaskType* pt, int dir)
{
    int hp,lp,newpr;
    list<ConnectionType*>::iterator ci;
    
    switch(dir)
    {
        case 0: // Moving down
            {
                hp=pt->Property()->Get_Priority()-1;
                lp=hp;

                for(ci=pt->Get_In_Edges_Begin();ci!=pt->Get_In_Edges_End();ci++)
                {
                    ConnectionType* pc=*ci;

                    TaskType* ptp=pc->Get_From();
                    if(ptp->Property()->Get_Priority()<lp) 
                        lp=ptp->Property()->Get_Priority();
                }
                if(pt->Get_Number_Of_In_Connection()==0) lp=0;
                newpr=lp+(int)((float)(hp-lp+1)*rand()/(RAND_MAX+1.0));
                return newpr;
            }
            break;
        case 1: // Moving up
            {
                lp=pt->Property()->Get_Priority()+1;
                hp=lp;

                for(ci=pt->Get_Out_Edges_Begin();ci!=pt->Get_Out_Edges_End();ci++)
                {
                    ConnectionType* pc=*ci;

                    TaskType* ptt=pc->Get_To();
                    if(ptt->Property()->Get_Priority()>hp)
                        hp=ptt->Property()->Get_Priority();
                }
                if(pt->Get_Number_Of_Out_Connection()==0) hp=taskidmap.size()-1;
                newpr=lp+(int)((float)(hp-lp+1)*rand()/(RAND_MAX+1.0));
                return newpr;
            }
            break;
    }
}

static void order_tasks_list (list<TaskType*>& task_list, list<TaskType*>&tl)
{
    list<TaskType*>::iterator otli;
    list<TaskType*>::iterator otask_listi;
    bool inserted;
    
    for(otli=tl.begin();otli!=tl.end();otli++)
    {
        TaskType *pt=*otli;
        inserted=false;
        
        for(otask_listi=task_list.begin();otask_listi!=task_list.end();otask_listi++)
        {
            TaskType *pt2=*otask_listi;
            float rel_deadline1, rel_deadline2;

            rel_deadline1=pt->Property()->Get_Global_Deadline();
            rel_deadline2=pt2->Property()->Get_Global_Deadline();
            
            if(rel_deadline1<rel_deadline2)
            {
                task_list.insert(otask_listi,pt);
                inserted=true;
                break;
            }
        }
        if(!inserted) task_list.push_back(pt);
    }
}

static float get_new_temperature (SA_parameters& sa, float temp)
{
    return sa.temp_const*temp;
}

static void unique_rand_init()
{
    rand_int_list.clear();
}

static int unique_rand (int size)
{
    while(1)
    {
        int rand_root=(int)(((float)size*rand())/(RAND_MAX+1.0));

        set<int>::iterator ii=rand_int_list.find(rand_root);
        if(ii==rand_int_list.end()) 
        {
            rand_int_list.insert(rand_root);
            return rand_root;
        }
        if(rand_int_list.size()==size) return rand_root;
            
    }
}
   
static bool check_utilization (Processor* pproc)
{
    float sum=0;
    list<TaskType*>::iterator ti;

    for(ti=pproc->Tasks_Begin();ti!=pproc->Tasks_End();ti++)
    {
        TaskType* pt=*ti;

        if(pt->Property()->Get_Global_Root()==pt)
            sum+=(pt->Property()->Get_Global_Exe_Time()/pt->Property()->Get_Global_Period());
    }
    return ((sum+0.01)<1.0);
}

static void algo_clear()
{
    taskidmap.clear();
    rand_int_list.clear();
}

